import contextlib
import enum
import logging
import os
import subprocess

from devtools import debug
from pydantic.v1 import (
    BaseModel,
    Field,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.model.common import (
    CommonCollectionParamters,
    LegacyParameters,
    PathParameters,
    StandardCollectionParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class SSXUserCollectionParameters(BaseModel):
    sub_sampling: float = Field(4, gt=0, lt=100)
    exp_time: float = Field(100e-6, gt=0, lt=1)
    take_pedestal: bool = Field(True)
    align_chip: bool = Field(True)

    class Config:
        extra: "ignore"


class SsxChipColletionTaskParameters(BaseModel):
    path_parameters: PathParameters
    common_parameters: CommonCollectionParamters
    collection_parameters: StandardCollectionParameters
    user_collection_parameters: SSXUserCollectionParameters
    legacy_parameters: LegacyParameters


class SsxChipCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SsxChipCollectionQueueEntry(BaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxChipCollectionQueueModel
    DATA_MODEL = SsxChipColletionTaskParameters
    NAME = "SSX Chip Collection (Lima1)"
    REQUIRES = ["point", "line", "no_shape", "chip", "mesh"]

    def __init__(self, view, data_model: SsxChipCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()
        params = self._data_model._task_data.user_collection_parameters
        self._data_model._task_data.collection_parameters.num_images = 400
        data_root_path, _ = self.get_data_path()
        self.take_pedestal()

        selected_regions = self._data_model._task_data.collection_parameters.selection
        selected_regions = selected_regions if selected_regions else [[0, 0]]

        for region in selected_regions:
            fname_prefix = self._data_model._task_data.path_parameters.prefix
            fname_prefix += f"_block_{region[0]}_{region[1]}_"

            HWR.beamline.diffractometer.set_phase("Centring", wait=True, timeout=120)
            # HWR.beamline.diffractometer.wait_ready()
            logging.getLogger("user_level_log").info(f"Acquiring {region} ...")

            if params.align_chip:
                logging.getLogger("user_level_log").info(f"Aligning block {region}")
                HWR.beamline.diffractometer.auto_align_ssx_block(region[0], region[1])
                logging.getLogger("user_level_log").info(f"Aligned block {region}")

            logging.getLogger("user_level_log").info("Preparing detector")

            HWR.beamline.detector.stop_acquisition()
            HWR.beamline.detector.prepare_acquisition(
                400, params.exp_time, data_root_path, fname_prefix
            )

            HWR.beamline.detector.wait_ready()

            HWR.beamline.detector.start_acquisition()
            logging.getLogger("user_level_log").info(
                "Detector ready, waiting for trigger ..."
            )
            HWR.beamline.diffractometer.wait_ready()

            logging.getLogger("user_level_log").info(f"Preparing data collection")
            HWR.beamline.diffractometer.set_phase(
                "DataCollection", wait=True, timeout=120
            )

            if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
                logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
                HWR.beamline.control.safshut_oh2.open()

            logging.getLogger("user_level_log").info(f"Scanning {region} ...")
            HWR.beamline.diffractometer.start_ssx_scan(params.sub_sampling)
            HWR.beamline.diffractometer.wait_ready()

            if HWR.beamline.control.safshut_oh2.state.name == "OPEN":
                HWR.beamline.control.safshut_oh2.close()
                logging.getLogger("user_level_log").info("shutter closed")

            HWR.beamline.detector.wait_ready()
            logging.getLogger("user_level_log").info(f"Acquired {region}")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
All queue entries inherits the baseclass BaseQueueEntry which in turn
inherits QueueEntryContainer. This makes it possible to arrange and
execute queue entries in a hierarchical manner.
"""

import copy
import logging
import sys
import time
import traceback
from collections import namedtuple

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects import autoprocessing
from mxcubecore.model import queue_model_objects
from mxcubecore.model.queue_model_enumerables import (
    CENTRING_METHOD,
    EXPERIMENT_TYPE,
)

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


status_list = ["SUCCESS", "WARNING", "FAILED", "SKIPPED", "RUNNING", "NOT_EXECUTED"]
QueueEntryStatusType = namedtuple("QueueEntryStatusType", status_list)
QUEUE_ENTRY_STATUS = QueueEntryStatusType(0, 1, 2, 3, 4, 5)


class QueueExecutionException(Exception):
    def __init__(self, message, origin):
        Exception.__init__(self, message, origin)
        self.message = message
        self.origin = origin
        if sys.exc_info()[0] is None:
            self.stack_trace = None
        else:
            self.stack_trace = traceback.format_exc()


class QueueAbortedException(QueueExecutionException):
    pass


class QueueSkipEntryException(QueueExecutionException):
    pass


class QueueEntryContainer(object):
    """
    A QueueEntryContainer has a list of queue entries, classes
    inheriting BaseQueueEntry, and a Queue object. The Queue object
    controls/handles the execution of the queue entries.
    """

    def __init__(self):
        object.__init__(self)
        self._queue_entry_list = []
        self._queue_controller = None
        self._parent_container = None

    def get_queue_entry_list(self):
        return self._queue_entry_list

    def enqueue(self, queue_entry):
        # A queue entry container has a QueueController object
        # which controls the execution of the tasks in the
        # container. QueueManagers are their own controller.
        # These are set in subclasses
        queue_entry.set_container(self)
        self._queue_entry_list.append(queue_entry)

    def dequeue(self, queue_entry):
        """
        Dequeues the QueueEntry <queue_entry> and returns the
        dequeued entry.

        Throws ValueError if the queue_entry is not in the queue.

        :param queue_entry: The queue entry to dequeue/remove.
        :type queue_entry: QueueEntry

        :returns: The dequeued entry.
        :rtype: QueueEntry
        """
        result = None
        index = None
        queue_entry.set_queue_controller(None)
        queue_entry.set_container(None)

        try:
            index = self._queue_entry_list.index(queue_entry)
        except ValueError:
            raise

        if index is not None:
            result = self._queue_entry_list.pop(index)

        log = logging.getLogger("queue_exec")
        msg = "dequeue called with: " + str(queue_entry)
        log.info(msg)

        return result

    def swap(self, queue_entry_a, queue_entry_b):
        """
        Swaps places between the two queue entries <queue_entry_a> and
        <queue_entry_b>.

        Throws a ValueError if one of the entries does not exist in the
        queue.

        :param queue_entry: Queue entry to swap
        :type queue_entry: QueueEntry

        :param queue_entry: Queue entry to swap
        :type queue_entry: QueueEntry
        """
        index_a = None
        index_b = None

        try:
            index_a = self._queue_entry_list.index(queue_entry_a)
        except ValueError:
            raise

        try:
            index_b = self._queue_entry_list.index(queue_entry_b)
        except ValueError:
            raise

        if (index_a is not None) and (index_b is not None):
            temp = self._queue_entry_list[index_a]
            self._queue_entry_list[index_a] = self._queue_entry_list[index_b]
            self._queue_entry_list[index_b] = temp

        log = logging.getLogger("queue_exec")
        msg = "swap called with: " + str(queue_entry_a) + ", " + str(queue_entry_b)
        log.info(msg)

        msg = "Queue is :" + str(self.get_queue_controller())
        log.info(msg)

    def set_queue_controller(self, queue_controller):
        """
        Sets the queue controller, the object that controls execution
        of this QueueEntryContainer.

        :param queue_controller: The queue controller object.
        :type queue_controller: QueueController
        """
        self._queue_controller = queue_controller

    def get_queue_controller(self):
        """
        :returns: The queue controller
        :type queue_controller: QueueController
        """
        return self._queue_controller

    def set_container(self, queue_entry_container):
        """
        Sets the parent queue entry to <queue_entry_container>

        :param queue_entry_container:
        :type queue_entry_container: QueueEntryContainer
        """
        self._parent_container = queue_entry_container

    def get_container(self):
        """
        :returns: The parent QueueEntryContainer.
        :rtype: QueueEntryContainer
        """
        return self._parent_container


class BaseQueueEntry(QueueEntryContainer):
    """
    Base class for queue entry objects. Defines the overall
    interface and behaviour for a queue entry.
    """

    def __init__(self, view=None, data_model=None, view_set_queue_entry=True):
        QueueEntryContainer.__init__(self)
        self._data_model = None
        self._view = None
        self.set_data_model(data_model)
        self.set_view(view, view_set_queue_entry)
        self._checked_for_exec = False
        self.status = QUEUE_ENTRY_STATUS.NOT_EXECUTED
        self.type_str = ""

    def is_failed(self):
        """Returns True if failed"""
        return self.status == QUEUE_ENTRY_STATUS.FAILED

    def enqueue(self, queue_entry):
        """
        Method inherited from QueueEntryContainer, a derived class
        should newer need to override this method.
        """
        queue_entry.set_queue_controller(self.get_queue_controller())
        super(BaseQueueEntry, self).enqueue(queue_entry)

    def set_data_model(self, data_model):
        """
        Sets the model node of this queue entry to <data_model>

        :param data_model: The data model node.
        :type data_model: TaskNode
        """
        self._data_model = data_model

    def get_data_model(self):
        """
        :returns: The data model of this queue entry.
        :rtype: TaskNode
        """
        return self._data_model

    def set_view(self, view, view_set_queue_entry=True):
        """
        Sets the view of this queue entry to <view>. Makes the
        corresponding bidirectional connection if view_set_queue_entry
        is set to True. Which is normally case, it can be useful with
        'uni-directional' connection in some rare cases.

        :param view: The view to associate with this entry
        :type view: ViewItem

        :param view_set_queue_entry: Bi- or uni-directional
                                     connection to view.
        :type view_set_queue_entry: bool
        """
        if view:
            self._view = view

            if view_set_queue_entry:
                view.set_queue_entry(self)

    def get_view(self):
        """
        :returns the view:
        :rtype: ViewItem
        """
        return self._view

    def is_enabled(self):
        """
        :returns: True if this item is enabled.
        :rtype: bool
        """
        return self._checked_for_exec

    def set_enabled(self, state):
        """
        Enables or disables this entry, controls if this item
        should be executed (enabled) or not (disabled)

        :param state: Enabled if state is True otherwise disabled.
        :type state: bool
        """
        self._checked_for_exec = state

    def execute(self):
        """
        Execute method, should be overriden my subclasses, defines
        the main body of the procedure to be performed when the entry
        is executed.

        The default executer calls excute on all child entries after
        this method but before post_execute.
        """
        msg = "Calling execute on: " + str(self)
        logging.getLogger("queue_exec").info(msg)

    def pre_execute(self):
        """
        Procedure to be done before execute.
        """
        msg = "Calling pre_execute on: " + str(self)
        logging.getLogger("queue_exec").info(msg)
        self.get_data_model().set_running(True)

    def post_execute(self):
        """
        Procedure to be done after execute, and execute of all
        children of this entry.
        """
        msg = "Calling post_execute on: " + str(self)
        logging.getLogger("queue_exec").info(msg)

        # view = self.get_view()
        # view.setHighlighted(True)
        # view.setOn(False)

        self.get_data_model().set_executed(True)
        self.get_data_model().set_running(False)
        self.get_data_model().set_enabled(False)
        self.set_enabled(False)

        # self._set_background_color()

    def _set_background_color(self):
        view = self.get_view()

        if self.get_data_model().is_executed():
            view.set_background_color(self.status + 1)
        else:
            view.set_background_color(0)
            # view.setBackgroundColor(widget_colors.WHITE)

    def stop(self):
        """
        Stops the execution of this entry, should free
        external resources, cancel all pending processes and so on.
        """
        self.get_view().setText(1, "Stopped")
        msg = "Calling stop on: " + str(self)
        logging.getLogger("queue_exec").info(msg)

    def handle_exception(self, ex):
        view = self.get_view()

        if view and isinstance(ex, QueueExecutionException):
            if ex.origin is self:
                # view.setBackgroundColor(widget_colors.LIGHT_RED)
                view.set_background_color(3)

    def __str__(self):
        info_str = "<%s object at %s> [" % (self.__class__.__name__, hex(id(self)))

        for entry in self._queue_entry_list:
            info_str += str(entry)

        return info_str + "]"

    def get_type_str(self):
        return self.type_str


class TaskGroupQueueEntry(BaseQueueEntry):
    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)

        self.interleave_task = None
        self.interleave_items = None
        self.interleave_sw_list = None
        self.interleave_stopped = None

    def execute(self):
        BaseQueueEntry.execute(self)
        task_model = self.get_data_model()
        gid = task_model.lims_group_id

        do_new_dc_group = True
        # Do not create a new data collection group if one already exists
        # or if the current task group contains a GenericWorkflowQueueEntry

        if gid:
            do_new_dc_group = False
        elif len(self._queue_entry_list) > 0:
            from mxcubecore.queue_entry.generic_workflow import (
                GenericWorkflowQueueEntry,
            )

            if isinstance(self._queue_entry_list[0], GenericWorkflowQueueEntry):
                do_new_dc_group = False

        init_ref_images = False
        if do_new_dc_group:
            # Creating a collection group with the current session id
            # and a dummy exepriment type OSC. The experiment type
            # will be updated when the collections are stored.
            if task_model.interleave_num_images:
                init_ref_images = task_model.interleave_num_images
                group_data = {
                    "sessionId": HWR.beamline.session.session_id,
                    "experimentType": "Collect - Multiwedge",
                }
            elif task_model.inverse_beam_num_images:
                init_ref_images = task_model.inverse_beam_num_images
                group_data = {
                    "sessionId": HWR.beamline.session.session_id,
                    "experimentType": "Collect - Multiwedge",
                }
            else:
                group_data = {
                    "sessionId": HWR.beamline.session.session_id,
                    "experimentType": "OSC",
                }

            sample_model = task_model.get_sample_node()
            task_model.get_parent()
            if sample_model.lims_container_location != -1:
                loc = sample_model.lims_container_location

                if isinstance(loc, str):
                    cell, puck = list(map(int, "2:2".split(":")))
                    loc = (cell - 1) * 3 + puck

                group_data["actualContainerSlotInSC"] = loc
            if sample_model.lims_sample_location != -1:
                group_data["actualSampleSlotInContainer"] = int(
                    sample_model.lims_sample_location
                )

            try:
                gid = HWR.beamline.lims._store_data_collection_group(group_data)
                self.get_data_model().lims_group_id = gid
            except Exception as ex:
                msg = (
                    "Could not create the data collection group"
                    + " in LIMS. Reason: "
                    + str(ex)
                )
                raise QueueExecutionException(msg, self)

        self.interleave_items = []
        if init_ref_images:
            # At first all children are gathered together and
            # checked if interleave is set. For this implementation
            # interleave is just possible for discreet data collections
            ref_num_images = 0
            children_data_model_list = self._data_model.get_children()

            for child_data_model in children_data_model_list:
                if isinstance(child_data_model, queue_model_objects.DataCollection):
                    if task_model.inverse_beam_num_images is not None:
                        child_data_model.acquisitions[
                            0
                        ].acquisition_parameters.num_images /= 2
                    num_images = child_data_model.acquisitions[
                        0
                    ].acquisition_parameters.num_images

                    if num_images > init_ref_images:
                        if num_images > ref_num_images:
                            ref_num_images = num_images
                        interleave_item = {}
                        child_data_model.set_experiment_type(
                            EXPERIMENT_TYPE.COLLECT_MULTIWEDGE
                        )
                        interleave_item["data_model"] = child_data_model
                        for queue_entry in self._queue_entry_list:
                            if queue_entry.get_data_model() == child_data_model:
                                interleave_item["queue_entry"] = queue_entry
                                interleave_item["tree_item"] = queue_entry.get_view()
                        self.interleave_items.append(interleave_item)

                        if task_model.inverse_beam_num_images is not None:
                            inverse_beam_item = copy.deepcopy(interleave_item)
                            inverse_beam_item["data_model"] = interleave_item[
                                "data_model"
                            ].copy()
                            inverse_beam_item["data_model"].acquisitions[
                                0
                            ].acquisition_parameters.osc_start += 180
                            inverse_beam_item["data_model"].acquisitions[
                                0
                            ].acquisition_parameters.first_image = (
                                interleave_item["data_model"]
                                .acquisitions[0]
                                .acquisition_parameters.first_image
                                + interleave_item["data_model"]
                                .acquisitions[0]
                                .acquisition_parameters.num_images
                            )
                            self.interleave_items.append(inverse_beam_item)
        if len(self.interleave_items) > 1:
            interleave_num_images = task_model.interleave_num_images
            self.interleave_task = gevent.spawn(
                self.execute_interleaved, ref_num_images, init_ref_images
            )
            self.interleave_task.join()

    def execute_interleaved(self, ref_num_images, interleave_num_images):
        task_model = self.get_data_model()

        if task_model.interleave_num_images:
            method_type = "interleave"
        elif task_model.inverse_beam_num_images:
            method_type = "inverse beam"

        logging.getLogger("queue_exec").info(
            "Preparing %s data collection" % method_type
        )

        for interleave_item in self.interleave_items:
            interleave_item["queue_entry"].set_enabled(False)
            interleave_item["tree_item"].set_checkable(False)
            interleave_item["data_model"].lims_group_id = (
                interleave_item["data_model"].get_parent().lims_group_id
            )
            cpos = (
                interleave_item["data_model"]
                .acquisitions[0]
                .acquisition_parameters.centred_position
            )
            # sample = interleave_item["data_model"].get_parent().get_parent()
            sample = interleave_item["data_model"].get_sample_node()
            empty_cpos = queue_model_objects.CentredPosition()
            param_list = queue_model_objects.to_collect_dict(
                interleave_item["data_model"],
                sample,
                cpos if cpos != empty_cpos else None,
            )
            # HWR.beamline.collect.prepare_interleave(
            #    interleave_item["data_model"], param_list
            # )

        self.interleave_sw_list = queue_model_objects.create_interleave_sw(
            self.interleave_items, ref_num_images, interleave_num_images
        )

        self._queue_controller.emit("queue_interleaved_started")
        for item_index, item in enumerate(self.interleave_sw_list):
            if not self.interleave_stopped:
                self.get_view().setText(
                    1,
                    "Subwedge %d:%d)"
                    % ((item_index + 1), len(self.interleave_sw_list)),
                )

                acq_par = (
                    self.interleave_items[item["collect_index"]]["data_model"]
                    .acquisitions[0]
                    .acquisition_parameters
                )

                acq_path_template = (
                    self.interleave_items[item["collect_index"]]["data_model"]
                    .acquisitions[0]
                    .path_template
                )

                acq_first_image = acq_par.first_image

                acq_par.first_image = item["sw_first_image"]
                acq_par.num_images = item["sw_actual_size"]
                acq_par.osc_start = item["sw_osc_start"]
                acq_par.in_interleave = (
                    acq_first_image,
                    acq_first_image + item["collect_num_images"] - 1,
                )
                self.interleave_items[item["collect_index"]]["queue_entry"].in_queue = (
                    item_index < (len(self.interleave_sw_list) - 1)
                )

                msg = "Executing %s collection (subwedge %d:%d, " % (
                    method_type,
                    (item_index + 1),
                    len(self.interleave_sw_list),
                )
                msg += "from %d to %d, " % (
                    acq_par.first_image,
                    acq_par.first_image + acq_par.num_images - 1,
                )
                msg += "osc start: %.2f, osc total range: %.2f)" % (
                    item["sw_osc_start"],
                    item["sw_osc_range"],
                )
                logging.getLogger("user_level_log").info(msg)

                try:
                    self.interleave_items[item["collect_index"]][
                        "queue_entry"
                    ].pre_execute()
                    self.interleave_items[item["collect_index"]][
                        "queue_entry"
                    ].execute()
                except Exception:
                    pass
                self.interleave_items[item["collect_index"]][
                    "queue_entry"
                ].post_execute()
                self.interleave_items[item["collect_index"]]["tree_item"].setText(
                    1,
                    "Subwedge %d:%d done"
                    % (item["collect_index"] + 1, item["sw_index"] + 1),
                )

                sig_data = {
                    "current_idx": item_index,
                    "item": item,
                    "nitems": len(self.interleave_sw_list),
                    "sw_size": interleave_num_images,
                }

                self._queue_controller.emit("queue_interleaved_sw_done", (sig_data,))

        if not self.interleave_stopped:
            logging.getLogger("queue_exec").info(
                "%s collection finished" % method_type.title()
            )
            self._queue_controller.emit("queue_interleaved_finished")

        self.interleave_task = None

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)

    def post_execute(self):
        BaseQueueEntry.post_execute(self)
        self.get_view().setText(1, "")

    def stop(self):
        BaseQueueEntry.stop(self)
        if self.interleave_task:
            self.interleave_stopped = True
            self.interleave_task.kill()
        self.get_view().setText(1, "Interleave stoped")


class SampleQueueEntry(BaseQueueEntry):
    """
    Defines the behaviour of sample queue entries. Mounting, launching centring
    and so on.
    """

    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)
        self.sample_centring_result = None

    def __getstate__(self):
        d = dict(self.__dict__)
        d["sample_centring_result"] = None
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)

    def execute(self):
        BaseQueueEntry.execute(self)
        log = logging.getLogger("queue_exec")
        sc_used = not self._data_model.free_pin_mode

        # Only execute samples with collections and when sample changer is used
        if len(self.get_data_model().get_children()) != 0 and sc_used:
            if HWR.beamline.diffractometer.in_plate_mode():
                return
            else:
                mount_device = HWR.beamline.sample_changer

            if mount_device is not None:
                log.info("Loading sample " + str(self._data_model.location))
                sample_mounted = mount_device.is_mounted_sample(
                    tuple(self._data_model.location)
                )
                if not sample_mounted:
                    self.sample_centring_result = gevent.event.AsyncResult()
                    try:
                        mount_sample(
                            self._data_model,
                            self.centring_done,
                            self.sample_centring_result,
                        )
                    except Exception as e:
                        self._view.setText(1, "Error loading")
                        msg = (
                            "Error loading sample, please check"
                            + " sample changer: "
                            + str(e)
                        )
                        log.error(msg)
                        self.status = QUEUE_ENTRY_STATUS.FAILED
                        if isinstance(e, QueueSkipEntryException):
                            raise
                        else:
                            raise QueueExecutionException(str(e), self)
                else:
                    log.info("Sample already mounted")
            else:
                msg = (
                    "SampleQueueItemPolicy does not have any "
                    + "sample changer hardware object, cannot "
                    + "mount sample"
                )
                log.info(msg)
            self.get_view().setText(1, "")

    def centring_done(self, success, centring_info):
        if not success:
            msg = (
                "Loop centring failed or was cancelled, " + "please continue manually."
            )
            logging.getLogger("user_level_log").warning(msg)
        self.sample_centring_result.set(centring_info)

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)

    def post_execute(self):
        BaseQueueEntry.post_execute(self)
        params = []

        # Start grouped processing, get information from each collection
        # and call autoproc with grouped processing option
        for child in self.get_data_model().get_children():
            for grand_child in child.get_children():
                if isinstance(grand_child, queue_model_objects.DataCollection):
                    xds_dir = grand_child.acquisitions[0].path_template.xds_dir
                    residues = grand_child.processing_parameters.num_residues
                    anomalous = grand_child.processing_parameters.anomalous
                    space_group = grand_child.processing_parameters.space_group
                    cell = grand_child.processing_parameters.get_cell_str()
                    inverse_beam = grand_child.acquisitions[
                        0
                    ].acquisition_parameters.inverse_beam

                    params.append(
                        {
                            "collect_id": grand_child.id,
                            "xds_dir": xds_dir,
                            "residues": residues,
                            "anomalous": anomalous,
                            "spacegroup": space_group,
                            "cell": cell,
                            "inverse_beam": inverse_beam,
                        }
                    )

        try:
            programs = HWR.beamline.collect["auto_processing"]
            autoprocessing.start(programs, "end_multicollect", params)
        except KeyError:
            pass

        self._set_background_color()
        self._view.setText(1, "")

    def _set_background_color(self):
        BaseQueueEntry._set_background_color(self)

    def get_type_str(self):
        return "Sample"


class BasketQueueEntry(BaseQueueEntry):
    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)


def mount_sample(data_model, centring_done_cb, async_result):
    HWR.beamline.sample_changer.trigger_progress_message("Loading sample")
    HWR.beamline.sample_view.clear_all()
    log = logging.getLogger("user_level_log")

    loc = data_model.location

    robot_action_dict = {
        "actionType": "LOAD",
        "containerLocation": loc[1],
        "dewarLocation": loc[0],
        "sampleBarcode": data_model.code,
        "sampleId": data_model.lims_id,
        "sessionId": HWR.beamline.session.session_id,
        "startTime": time.strftime("%Y-%m-%d %H:%M:%S"),
    }

    if not HWR.beamline.sample_changer.load(sample=data_model.loc_str, wait=True):
        raise QueueSkipEntryException("Sample changer could not load sample", "")

    robot_action_dict["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")

    if HWR.beamline.sample_changer.has_loaded_sample():
        robot_action_dict["status"] = "SUCCESS"
    else:
        robot_action_dict["message"] = "Sample was not loaded"
        robot_action_dict["status"] = "ERROR"

    HWR.beamline.lims.store_robot_action(robot_action_dict)

    if not HWR.beamline.sample_changer.has_loaded_sample():
        HWR.beamline.sample_changer.trigger_progress_message("Sample not loaded")
        raise QueueSkipEntryException("Sample not loaded", "")
    else:
        HWR.beamline.sample_changer.trigger_progress_message("Sample loaded")
        dm = HWR.beamline.diffractometer
        centring_method = HWR.beamline.queue_manager.centring_method

        if centring_method != CENTRING_METHOD.NONE:
            try:
                dm.connect("centringAccepted", centring_done_cb)

                if centring_method == CENTRING_METHOD.MANUAL:
                    log.warning(
                        "Manual centring used, waiting for" + " user to center sample"
                    )
                    dm.start_centring_method(dm.MANUAL3CLICK_MODE)
                elif centring_method == CENTRING_METHOD.LOOP:
                    dm.start_centring_method(dm.C3D_MODE)
                    log.warning(
                        "Centring in progress. Please save"
                        + " the suggested centring or re-center"
                    )
                elif centring_method == CENTRING_METHOD.FULLY_AUTOMATIC:
                    log.info("Centring sample, please wait.")
                    dm.start_centring_method(dm.C3D_MODE)
                else:
                    dm.start_centring_method(dm.MANUAL3CLICK_MODE)

                HWR.beamline.sample_changer.trigger_progress_message("Centring !")
                centring_result = async_result.get()

                if centring_result["valid"]:
                    HWR.beamline.sample_changer.trigger_progress_message(
                        "Centring done !"
                    )
                    log.info("Centring saved")
                else:
                    HWR.beamline.sample_changer.trigger_progress_message(
                        "Centring failed !"
                    )
                    if centring_method == CENTRING_METHOD.FULLY_AUTOMATIC:
                        raise QueueSkipEntryException(
                            "Could not center sample, skipping", ""
                        )
                    else:
                        raise RuntimeError("Could not center sample")
            except Exception:
                logging.getLogger("HWR").exception("")
            finally:
                dm.disconnect("centringAccepted", centring_done_cb)


class DelayQueueEntry(BaseQueueEntry):
    def __init__(self, view=None, data_model=None):
        BaseQueueEntry.__init__(self, view, data_model)

    def execute(self):
        BaseQueueEntry.execute(self)
        delay = self.get_data_model().delay
        logging.getLogger("HWR").debug("Execute Delay entry, delay =  %s" % delay)
        time.sleep(delay)


def center_before_collect(view, dm, queue, sample_view):
    view.setText(1, "Waiting for input")
    log = logging.getLogger("user_level_log")

    log.info("Please select, or center on a new position and press continue.")

    queue.pause(True)
    pos, shape = None, None

    if len(sample_view.get_selected_shapes()):
        shape = sample_view.get_selected_shapes()[0]
        pos = shape.mpos()
    else:
        msg = "No centred position selected, using current position."
        log.info(msg)

        # Create a centred postions of the current postion
        pos = dm.get_positions()
        shape = sample_view.add_shape_from_mpos([pos], (0, 0), "P")

    view(1, "Centring completed")
    log.info("Centring completed")

    return queue_model_objects.CentredPosition(pos), shape



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractXRFSpectrum import AbstractXRFSpectrum

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "Task"


class EMBLXRFSpectrum(AbstractXRFSpectrum):
    def __init__(self, name):
        AbstractXRFSpectrum.__init__(self, name)

        self.ready_event = None
        self.spectrum_running = None
        self.spectrum_info = None
        self.config_filename = None
        self.spectrum_data = None
        self.mca_calib = None

        self.chan_spectrum_status = None
        self.chan_spectrum_consts = None
        self.chan_scan_error = None
        self.cmd_spectrum_start = None
        self.cmd_adjust_transmission = None

    def init(self):
        self.ready_event = gevent.event.Event()

        self.cmd_spectrum_start = self.get_command_object("cmdSpectrumStart")
        self.cmd_adjust_transmission = self.get_command_object("cmdAdjustTransmission")

        self.chan_spectrum_status = self.get_channel_object("chanSpectrumStatus")
        self.chan_spectrum_status.connect_signal("update", self.spectrum_status_update)
        self.chan_spectrum_consts = self.get_channel_object("chanSpectrumConsts")

        self.chan_scan_error = self.get_channel_object("chanSpectrumError")
        self.chan_scan_error.connect_signal("update", self.scan_error_update)

        self.config_filename = self.get_property("configFile")
        self.write_in_raw_data = False
        self.mca_calib = [
            x / 1000.0 for x in self.chan_spectrum_consts.get_value()
        ]  # converted from eV to keV

    def can_spectrum(self):
        """
        Returns True if one can run spectrum command
        :return:
        """
        return True

    def execute_spectrum_command(self, count_sec, filename, adjust_transmission=True):
        """Sends start command"""
        try:
            self.cmd_spectrum_start((count_sec, adjust_transmission))
        except Exception:
            logging.getLogger().exception("XRFSpectrum: problem in starting spectrum")
            self.emit(
                "xrfSpectrumStatusChanged", ("Error problem in starting spectrum",)
            )
            self.spectrum_command_aborted()

    def spectrum_status_update(self, status):
        """Controls execution"""
        if self.spectrum_running:
            if status == "scaning":
                logging.getLogger("HWR").info("XRF spectrum in progress...")
            elif status == "ready":
                if self.spectrum_running:
                    self.spectrum_data = list(self.cmd_spectrum_start.get())
                    # self.mca_calib = self.chan_spectrum_consts.get_value()[::-1]
                    self.spectrum_command_finished()
                    logging.getLogger("HWR").info("XRF spectrum finished")
            elif status == "aborting":
                if self.spectrum_running:
                    self.spectrum_command_aborted()
                    logging.getLogger("HWR").info("XRF spectrum aborted!")
            elif status == "error":
                self.spectrum_command_failed()
                logging.getLogger("HWR").error("XRF spectrum failed!")

    def cancel_spectrum(self, *args):
        """Cancels acquisition"""
        if self.spectrum_running:
            self.ready_event.set()

    def adjust_transmission(self):
        """Adjusts transmission before executing XRF spectrum"""
        if self.cmd_adjust_transmission is not None:
            self.cmd_adjust_transmission()

    def scan_error_update(self, error_msg):
        """Prints error message

        :param error_msg: error message
        :type error_msg: str
        :return: None
        """
        if len(error_msg) > 0:
            # and self.spectrum_running:
            logging.getLogger("GUI").error("XRF spectrum: %s" % error_msg)



"""Bliss session and tools for sending the scan data for plotting.
Emits new_plot, plot_data and plot_end.
"""

import itertools

import gevent
import numpy
from bliss.config import static
from bliss.data.node import (
    DataNode,
    get_or_create_node,
)

from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


def all_equal(iterable):
    """Check for same number of points on each line"""
    grp = itertools.groupby(iterable)
    return next(grp, True) and not next(grp, False)


class Bliss(HardwareObject):
    """Bliss class"""

    def __init__(self, *args):
        HardwareObject.__init__(self, *args)
        self.__scan_data = {}

    def init(self, *args):
        """Initialis the bliss session"""
        cfg = static.get_config()
        session = cfg.get(self.get_property("session"))

        session.setup(self.__dict__, verbose=True)

        self.__scan_data = dict()

    def __on_scan_new(self, scan_info):
        """New scan. Emit new_plot.
        Args:
            scan_info(dict): Contains SCAN_INFO dictionary from bliss
        """
        scan_id = scan_info["scan_nb"]
        self.__scan_data[scan_id] = list()

        if not scan_info["save"]:
            scan_info["root_path"] = "<no file>"

        self.emit(
            "new_plot",
            {
                "id": scan_info["scan_nb"],
                "title": scan_info["title"],
                "labels": scan_info["labels"],
            },
        )

    def __on_scan_data(self, scan_info, data):
        """Retrieve the scan data. Emit plot_data.
        Args:
            scan_info (dict): SCAN_INFO dictionary from bliss
            data (numpy array): data from bliss
        """

        scan_id = scan_info["scan_nb"]
        new_data = numpy.column_stack([data[name] for name in scan_info["labels"]])
        self.__scan_data[scan_id].append(new_data)
        self.emit(
            "plot_data",
            {
                "id": scan_id,
                "data": numpy.concatenate(self.__scan_data[scan_id]).tolist(),
            },
        )

    def __on_scan_end(self, scan_info):
        """Retrieve remaining data at the end of the scan. Emit plot_end.
        Args:
            scan_info (int): ID of the scan
        """
        scan_id = scan_info["scan_nb"]
        self.emit(
            "plot_end",
            {
                "id": scan_id,
                "data": numpy.concatenate(self.__scan_data[scan_id]).tolist(),
            },
        )
        del self.__scan_data[scan_id]



import logging
import time

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.Command.Tango import DeviceProxy


class PX1TangoLight(HardwareObject):
    def __init__(self, name):
        super().__init__(name)
        self.currentState = "unknown"

    def init(self):
        # self.tangoname = self.
        self.attrchan = self.get_channel_object("attributeName")
        self.attrchan.connect_signal("update", self.value_changed)

        self.attrchan.connect_signal("connected", self._set_ready)
        self.attrchan.connect_signal("disconnected", self._set_ready)
        self.set_in = self.get_command_object("set_in")
        self.set_in.connect_signal("connected", self._set_ready)
        self.set_in.connect_signal("disconnected", self._set_ready)
        self.set_out = self.get_command_object("set_out")

        self.px1env_hwo = self.get_object_by_role("px1environment")
        self.light_hwo = self.get_object_by_role("intensity")
        self.zoom_hwo = self.get_object_by_role("zoom")

        self.connect(self.zoom_hwo, "predefinedPositionChanged", self.zoom_changed)

        self._set_ready()
        try:
            self.inversed = self.get_property("inversed")
        except Exception:
            self.inversed = False

        if self.inversed:
            self.states = ["in", "out"]
        else:
            self.states = ["out", "in"]

    def _set_ready(self):
        self.set_is_ready(self.attrchan.is_connected())

    def connect_notify(self, signal):
        if self.is_ready():
            self.value_changed(self.attrchan.get_value())

    def value_changed(self, value):
        self.currentState = value

        if value:
            self.currentState = self.states[1]
        else:
            self.currentState = self.states[0]

        self.emit("wagoStateChanged", (self.currentState,))

    def getWagoState(self):
        return self.currentState

    def wagoIn(self):
        self.setIn()

    def wagoOut(self):
        self.setOut()

    def setIn(self):
        if not self.px1env_hwo.isPhaseVisuSample():
            self.px1env_hwo.gotoSampleViewPhase()
            start_phase = time.time()
            while not self.px1env_hwo.isPhaseVisuSample():
                time.sleep(0.1)
                if time.time() - start_phase > 20:
                    break

        self.adjustLightLevel()

    def setOut(self):
        self._set_ready()
        if self.is_ready():
            if self.inversed:
                self.set_in()
            else:
                self.light_hwo.set_value(0)
                self.set_out()

    def zoom_changed(self, position_name, value):
        if self.currentState == "in":
            logging.getLogger("HWR").debug(
                "Zoom changed. and light is in. setting light level"
            )
            self.adjustLightLevel()

    def adjustLightLevel(self):
        if self.zoom_hwo is None or self.light_hwo is None:
            return

        props = self.zoom_hwo.getCurrentPositionProperties()

        try:
            if "lightLevel" in props.keys():
                light_level = float(props["lightLevel"])
                light_current = self.light_hwo.get_value()
                if light_current != light_level:
                    logging.getLogger("HWR").debug(
                        "Setting light level to %s" % light_level
                    )
                    self.light_hwo.set_value(light_level)
        except Exception:
            logging.getLogger("HWR").debug("Cannot set light level")



import logging
import math
import os
import tempfile
import time

import gevent.event
import numpy
from scipy import optimize

try:
    import lucid3 as lucid
except ImportError:
    try:
        import lucid
    except ImportError:
        logging.warning(
            "Could not find autocentring library, automatic centring is disabled"
        )


def multiPointCentre(z, phis):
    def fitfunc(p, x):
        return p[0] * numpy.sin(x + p[1]) + p[2]

    def errfunc(p, x, y):
        return fitfunc(p, x) - y

    # The function call returns tuples of varying length
    result = optimize.leastsq(errfunc, [1.0, 0.0, 0.0], args=(phis, z))
    return result[0]


USER_CLICKED_EVENT = None
CURRENT_CENTRING = None
SAVED_INITIAL_POSITIONS = {}
READY_FOR_NEXT_POINT = gevent.event.Event()
NUM_CENTRING_ROUNDS = 1


class CentringMotor:
    def __init__(self, motor, reference_position=None, direction=1, units="mm"):
        self.motor = motor
        self.direction = direction
        self.reference_position = reference_position
        self.units = units.lower()

        self._scale = 1.0  # mm or deg

        if units == "micron" or units == "microns":
            self._scale = 1000.0

    def mm_to_units(self, mm_dist):
        return mm_dist * self._scale

    def units_to_mm(self, mm_dist):
        return mm_dist / self._scale

    def __getattr__(self, attr):
        # delegate to motor object
        if attr.startswith("__"):
            raise AttributeError(attr)
        else:
            return getattr(self.motor, attr)


def prepare(centring_motors_dict):
    logging.debug("Preparing for centring")

    global SAVED_INITIAL_POSITIONS

    if CURRENT_CENTRING and not CURRENT_CENTRING.ready():
        end()

    global USER_CLICKED_EVENT
    global READY_FOR_NEXT_POINT
    USER_CLICKED_EVENT = gevent.event.AsyncResult()
    READY_FOR_NEXT_POINT = gevent.event.Event()

    motors_to_move = dict()
    for m in centring_motors_dict.values():
        if m.reference_position is not None:
            motors_to_move[m.motor] = m.reference_position
    move_motors(motors_to_move)

    SAVED_INITIAL_POSITIONS = dict(
        [(m.motor, m.motor.get_value()) for m in centring_motors_dict.values()]
    )

    phi = centring_motors_dict["phi"]
    phiy = centring_motors_dict["phiy"]
    sampx = centring_motors_dict["sampx"]
    sampy = centring_motors_dict["sampy"]
    phiz = centring_motors_dict["phiz"]

    return phi, phiy, phiz, sampx, sampy


def start(
    centring_motors_dict,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    chi_angle=0,
    n_points=3,
):
    global CURRENT_CENTRING

    phi, phiy, phiz, sampx, sampy = prepare(centring_motors_dict)

    CURRENT_CENTRING = gevent.spawn(
        center,
        phi,
        phiy,
        phiz,
        sampx,
        sampy,
        pixelsPerMm_Hor,
        pixelsPerMm_Ver,
        beam_xc,
        beam_yc,
        chi_angle,
        n_points,
    )
    return CURRENT_CENTRING


def start_plate(
    centring_motors_dict,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    plate_vertical,
    chi_angle=0,
    n_points=3,
    phi_range=10,
    lim_pos=314.0,
):
    global CURRENT_CENTRING

    plateTranslation = centring_motors_dict["plateTranslation"]
    centring_motors_dict.pop("plateTranslation")
    phi, phiy, phiz, sampx, sampy = prepare(centring_motors_dict)

    phi.set_value(lim_pos)

    CURRENT_CENTRING = gevent.spawn(
        centre_plate,
        phi,
        phiy,
        phiz,
        sampx,
        sampy,
        plateTranslation,
        pixelsPerMm_Hor,
        pixelsPerMm_Ver,
        beam_xc,
        beam_yc,
        plate_vertical,
        chi_angle,
        n_points,
        phi_range,
    )
    return CURRENT_CENTRING


def start_plate_1_click(
    centring_motors_dict,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    plate_vertical,
    phi_min,
    phi_max,
    n_points=10,
):
    global CURRENT_CENTRING

    # plateTranslation = centring_motors_dict["plateTranslation"]
    # centring_motors_dict.pop("plateTranslation")

    # phi, phiy,phiz, sampx, sampy = prepare(centring_motors_dict)

    phi = centring_motors_dict["phi"]
    phiy = centring_motors_dict["phiy"]
    sampx = centring_motors_dict["sampx"]
    sampy = centring_motors_dict["sampy"]
    phiz = centring_motors_dict["phiz"]

    # phi.set_value(phi_min)
    plate_vertical()

    CURRENT_CENTRING = gevent.spawn(
        centre_plate1Click,
        phi,
        phiy,
        phiz,
        sampx,
        sampy,
        pixelsPerMm_Hor,
        pixelsPerMm_Ver,
        beam_xc,
        beam_yc,
        plate_vertical,
        phi_min,
        phi_max,
        n_points,
    )

    return CURRENT_CENTRING


def centre_plate1Click(
    phi,
    phiy,
    phiz,
    sampx,
    sampy,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    plate_vertical,
    phi_min,
    phi_max,
    n_points,
):

    global USER_CLICKED_EVENT

    try:
        i = 0
        previous_click_x = 99999
        previous_click_y = 99999
        dx = 99999
        dy = 99999

        # while i < n_points and (dx > 3 or dy > 3) :
        # NBNB is this temporary or permanent?
        while (
            True
        ):  # it is now a while true loop that can be interrupted at any time by the save button, to allow user to have a 1 click centring as precise as he wants (see HutchMenuBrick)
            USER_CLICKED_EVENT = gevent.event.AsyncResult()
            try:
                x, y = USER_CLICKED_EVENT.get()
            except Exception:
                raise RuntimeError("Aborted while waiting for point selection")

            # Move to beam
            phiz.set_value_relative((y - beam_yc) / float(pixelsPerMm_Ver))
            phiy.set_value_relative(-(x - beam_xc) / float(pixelsPerMm_Hor))

            # Distance to previous click to end centring if it converges
            dx = abs(previous_click_x - x)
            dy = abs(previous_click_y - y)
            previous_click_x = x
            previous_click_y = y

            # Alterning between phi min and phi max to gradually converge to the
            # centring point
            if i % 2 == 0:
                phi_min = (
                    phi.get_value()
                )  # in case the phi range sent us to a position where sample is invisible, if user moves phi, this modifications is saved for future moves
                phi.set_value(phi_max)
            else:
                phi_max = (
                    phi.get_value()
                )  # in case the phi range sent us to a position where sample is invisible, if user moves phi, this modifications is saved for future moves
                phi.set_value(phi_min)

            READY_FOR_NEXT_POINT.set()
            i += 1
    except Exception:
        logging.exception("Exception while centring")
        move_motors(SAVED_INITIAL_POSITIONS)
        raise RuntimeError("Exception while centring")

    plate_vertical()

    centred_pos = SAVED_INITIAL_POSITIONS.copy()

    centred_pos.update(
        {sampx.motor: float(sampx.get_value()), sampy.motor: float(sampy.get_value())}
    )

    return centred_pos


def centre_plate(
    phi,
    phiy,
    phiz,
    sampx,
    sampy,
    plateTranslation,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    plate_vertical,
    chi_angle,
    n_points,
    phi_range=40,
):
    global USER_CLICKED_EVENT
    X, Y, phi_positions = [], [], []

    phi_angle = phi_range / (n_points - 1)

    try:
        i = 0
        while i < n_points:
            try:
                x, y = USER_CLICKED_EVENT.get()
            except Exception:
                raise RuntimeError("Aborted while waiting for point selection")
            USER_CLICKED_EVENT = gevent.event.AsyncResult()
            X.append(x / float(pixelsPerMm_Hor))
            Y.append(y / float(pixelsPerMm_Ver))
            phi_positions.append(phi.direction * math.radians(phi.get_value()))
            if i != n_points - 1:
                phi.set_value_relative(phi.direction * phi_angle, timeout=None)
            READY_FOR_NEXT_POINT.set()
            i += 1
    except Exception:
        logging.exception("Exception while centring")
        move_motors(SAVED_INITIAL_POSITIONS)
        raise

    # logging.info("X=%s,Y=%s", X, Y)
    chi_angle = math.radians(chi_angle)
    chiRotMatrix = numpy.matrix(
        [
            [math.cos(chi_angle), -math.sin(chi_angle)],
            [math.sin(chi_angle), math.cos(chi_angle)],
        ]
    )
    Z = chiRotMatrix * numpy.matrix([X, Y])
    z = Z[1]
    avg_pos = Z[0].mean()

    r, a, offset = multiPointCentre(numpy.array(z).flatten(), phi_positions)
    dy = r * numpy.sin(a)
    dx = r * numpy.cos(a)

    d = chiRotMatrix.transpose() * numpy.matrix([[avg_pos], [offset]])

    d_horizontal = d[0] - (beam_xc / float(pixelsPerMm_Hor))
    d_vertical = d[1] - (beam_yc / float(pixelsPerMm_Ver))

    centred_pos = SAVED_INITIAL_POSITIONS.copy()
    centred_pos.update(
        {
            sampx.motor: float(sampx.get_value() + sampx.direction * dx),
            sampy.motor: float(sampy.get_value() + sampy.direction * dy),
            phiz.motor: (
                float(phiz.get_value() + phiz.direction * d_vertical[0, 0])
                if phiz.__dict__.get("reference_position") is None
                else phiz.reference_position
            ),
            phiy.motor: (
                float(phiy.get_value() + phiy.direction * d_horizontal[0, 0])
                if phiy.__dict__.get("reference_position") is None
                else phiy.reference_position
            ),
        }
    )

    move_motors(centred_pos)
    plate_vertical()
    """
  try:
    x, y = USER_CLICKED_EVENT.get()
  except:
    raise RuntimeError("Aborted while waiting for point selection")
  USER_CLICKED_EVENT = gevent.event.AsyncResult()
  y_offset = -(y-beam_yc)  / float(pixelsPerMm_Ver)
  plateTranslation.set_value_relative(y_offset)
  """

    return centred_pos


def ready(motor_list):
    return all([m.is_ready() for m in motor_list])


def wait_ready(motor_positions_dict, timeout=None):
    with gevent.Timeout(timeout):
        while not ready(motor_positions_dict.keys()):
            time.sleep(0.1)


def move_motors(motor_positions_dict):
    if not motor_positions_dict:
        return

    wait_ready(motor_positions_dict, timeout=30)

    for motor, position in motor_positions_dict.items():
        motor.set_value(position)

    wait_ready(motor_positions_dict, timeout=60)


def user_click(x, y, wait=False):
    READY_FOR_NEXT_POINT.clear()
    USER_CLICKED_EVENT.set((x, y))
    if wait:
        READY_FOR_NEXT_POINT.wait()


def center(
    phi,
    phiy,
    phiz,
    sampx,
    sampy,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    chi_angle,
    n_points,
    phi_range=180,
):
    global USER_CLICKED_EVENT
    X, Y, phi_positions = [], [], []

    phi_angle = phi_range / (n_points - 1)

    try:
        i = 0
        while i < n_points:
            try:
                x, y = USER_CLICKED_EVENT.get()
            except Exception:
                raise RuntimeError("Aborted while waiting for point selection")
            USER_CLICKED_EVENT = gevent.event.AsyncResult()
            X.append(x / float(pixelsPerMm_Hor))
            Y.append(y / float(pixelsPerMm_Ver))
            phi_positions.append(phi.direction * math.radians(phi.get_value()))
            if i != n_points - 1:
                phi.set_value_relative(phi.direction * phi_angle, timeout=10)
            READY_FOR_NEXT_POINT.set()
            i += 1
    except Exception:
        logging.exception("Exception while centring")
        move_motors(SAVED_INITIAL_POSITIONS)
        READY_FOR_NEXT_POINT.set()
        raise RuntimeError("Exception while centring")

    # logging.info("X=%s,Y=%s", X, Y)
    chi_angle = math.radians(chi_angle)
    chiRotMatrix = numpy.matrix(
        [
            [math.cos(chi_angle), -math.sin(chi_angle)],
            [math.sin(chi_angle), math.cos(chi_angle)],
        ]
    )
    Z = chiRotMatrix * numpy.matrix([X, Y])
    z = Z[1]
    avg_pos = Z[0].mean()

    r, a, offset = multiPointCentre(numpy.array(z).flatten(), phi_positions)
    dy = r * numpy.sin(a)
    dx = r * numpy.cos(a)

    d = chiRotMatrix.transpose() * numpy.matrix([[avg_pos], [offset]])

    d_horizontal = d[0] - (beam_xc / float(pixelsPerMm_Hor))
    d_vertical = d[1] - (beam_yc / float(pixelsPerMm_Ver))

    phi_pos = math.radians(phi.direction * phi.get_value())

    centred_pos = SAVED_INITIAL_POSITIONS.copy()
    centred_pos.update(
        {
            sampx.motor: float(sampx.get_value() + sampx.direction * dx),
            sampy.motor: float(sampy.get_value() + sampy.direction * dy),
            phiz.motor: (
                float(phiz.get_value() + phiz.direction * d_vertical[0, 0])
                if phiz.__dict__.get("reference_position") is None
                else phiz.reference_position
            ),
            phiy.motor: (
                float(phiy.get_value() + phiy.direction * d_horizontal[0, 0])
                if phiy.__dict__.get("reference_position") is None
                else phiy.reference_position
            ),
        }
    )

    return centred_pos


def end(centred_pos=None):
    if centred_pos is None:
        centred_pos = CURRENT_CENTRING.get()
    try:
        move_motors(centred_pos)
    except Exception:
        READY_FOR_NEXT_POINT.set()
        move_motors(SAVED_INITIAL_POSITIONS)
        raise RuntimeError("Centring aborted")


def start_auto(
    sample_view,
    centring_motors_dict,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    chi_angle=0,
    n_points=3,
    msg_cb=None,
    new_point_cb=None,
):
    global CURRENT_CENTRING

    phi, phiy, phiz, sampx, sampy = prepare(centring_motors_dict)

    CURRENT_CENTRING = gevent.spawn(
        auto_center,
        sample_view,
        phi,
        phiy,
        phiz,
        sampx,
        sampy,
        pixelsPerMm_Hor,
        pixelsPerMm_Ver,
        beam_xc,
        beam_yc,
        chi_angle,
        n_points,
        msg_cb,
        new_point_cb,
    )
    return CURRENT_CENTRING


def find_loop(sample_view, pixelsPerMm_Hor, chi_angle, msg_cb, new_point_cb):
    snapshot_filename = os.path.join(
        tempfile.gettempdir(), "mxcube_sample_snapshot.png"
    )
    sample_view.save_snapshot(snapshot_filename, overlay=False, bw=True)

    # Lucid does not accept 0 degree rotation and
    # has a reference frame that is reversed to the one used
    # in MXCuBE
    if chi_angle == 0:
        chi_angle = None
    else:
        chi_angle = -chi_angle

    info, x, y = lucid.find_loop(
        snapshot_filename, rotation=chi_angle, debug=False, IterationClosing=6
    )

    try:
        x = float(x)
        y = float(y)
    except Exception:
        return -1, -1

    if callable(msg_cb):
        msg_cb("Loop found: %s (%d, %d)" % (info, x, y))
    if callable(new_point_cb):
        new_point_cb((x, y))

    return x, y


def auto_center(
    sample_view,
    phi,
    phiy,
    phiz,
    sampx,
    sampy,
    pixelsPerMm_Hor,
    pixelsPerMm_Ver,
    beam_xc,
    beam_yc,
    chi_angle,
    n_points,
    msg_cb,
    new_point_cb,
):
    imgWidth = sample_view.camera.get_width()
    imgHeight = sample_view.camera.get_height()

    # check if loop is there at the beginning
    i = 0
    while -1 in find_loop(
        sample_view, pixelsPerMm_Hor, chi_angle, msg_cb, new_point_cb
    ):
        phi.set_value_relative(90)
        i += 1
        if i > 4:
            if callable(msg_cb):
                msg_cb("No loop detected, aborting")
            return

    for k in range(NUM_CENTRING_ROUNDS):
        if callable(msg_cb):
            msg_cb("Doing automatic centring")

        centring_greenlet = gevent.spawn(
            center,
            phi,
            phiy,
            phiz,
            sampx,
            sampy,
            pixelsPerMm_Hor,
            pixelsPerMm_Ver,
            beam_xc,
            beam_yc,
            chi_angle,
            n_points,
        )

        for a in range(n_points):
            x, y = find_loop(
                sample_view, pixelsPerMm_Hor, chi_angle, msg_cb, new_point_cb
            )
            # logging.info("in autocentre, x=%f, y=%f",x,y)
            if x < 0 or y < 0:
                for i in range(1, 18):
                    # logging.info("loop not found - moving back %d" % i)
                    phi.set_value_relative(5)
                    x, y = find_loop(
                        sample_view,
                        pixelsPerMm_Hor,
                        chi_angle,
                        msg_cb,
                        new_point_cb,
                    )
                    if -1 in (x, y):
                        continue
                    if x >= 0:
                        if y < imgHeight / 2:
                            y = 0
                            if callable(new_point_cb):
                                new_point_cb((x, y))
                            user_click(x, y, wait=True)
                            break
                        else:
                            y = imgHeight
                            if callable(new_point_cb):
                                new_point_cb((x, y))
                            user_click(x, y, wait=True)
                            break
                if -1 in (x, y):
                    centring_greenlet.kill()
                    raise RuntimeError("Could not centre sample automatically.")
                phi.set_value_relative(-i * 5)
            else:
                user_click(x, y, wait=True)

        centred_pos = centring_greenlet.get()
        end(centred_pos)

    return centred_pos



import sys

from .Component import Component

try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen


class Sample(Component):
    """
    Entity class holding state of an individual sample or sample slot (an empty sample location).
    """

    # Common properties
    __HOLDER_LENGTH_PROPERTY__ = "Length"
    __IMAGE_URL_PROPERTY__ = "Image"
    __IMAGE_X_PROPERTY__ = "X"
    __IMAGE_Y_PROPERTY__ = "Y"
    __INFO_URL_PROPERTY__ = "Info"

    def __init__(self, container, address, scannable):
        super(Sample, self).__init__(container, address, scannable)
        self.properties = {}
        self.loaded = False
        self._has_been_loaded = False
        self._leaf = True

    #########################           PUBLIC           #########################

    def is_loaded(self):
        """
        Returns if the sample is currently loaded for data collection
        :rtype: bool
        """
        return self.loaded

    def has_been_loaded(self):
        """
        Returns if the sample has already beenloaded for data collection
        :rtype: bool
        """
        return self._has_been_loaded

    def get_properties(self):
        """
        Returns a dictionary with sample changer specific sample properties
        :rtype: dictionary
        """
        return self.properties

    def has_property(self, name):
        """
        Returns true if a property is defined
        :rtype: bool
        """
        return name in self.properties

    def get_property(self, name):
        """
        Returns a given property or None if not defined
        :rtype: object
        """
        if not self.has_property(name):
            return None

        return self.properties[name]

    def fetch_image(self):
        try:
            if self.has_property(self.__IMAGE_URL_PROPERTY__):
                img_url = self.get_property(self.__IMAGE_URL_PROPERTY__)
                if len(img_url) == 0:
                    return None

                f = urlopen(img_url)
                img = f.read()
                return img
        except Exception:
            print((sys.exc_info()[1]))

    def clear_info(self):
        Component.clear_info(self)
        changed = False
        if self.loaded:
            self.loaded = False
            changed = True
        if self._has_been_loaded:
            self._has_been_loaded = False
            changed = True
        if changed:
            self._set_dirty()

    # Common properties
    def get_holder_length(self):
        return self.get_property(self.__HOLDER_LENGTH_PROPERTY__)

    def _set_holder_length(self, value):
        self._set_property(self.__HOLDER_LENGTH_PROPERTY__, value)

    def _set_image_x(self, value):
        self._set_property(self.__IMAGE_X_PROPERTY__, value)

    def get_image_x(self):
        return self.get_property(self.__IMAGE_X_PROPERTY__)

    def _set_image_y(self, value):
        self._set_property(self.__IMAGE_Y_PROPERTY__, value)

    def get_image_y(self):
        return self.get_property(self.__IMAGE_Y_PROPERTY__)

    def _set_image_url(self, value):
        if (value is not None) and (value.startswith("http://")):
            value = "https://" + value[7]
        self._set_property(self.__IMAGE_URL_PROPERTY__, value)

    def get_image_url(self):
        return self.get_property(self.__IMAGE_URL_PROPERTY__)

    def _set_info_url(self, value):
        self._set_property(self.__INFO_URL_PROPERTY__, value)

    def get_info_url(self):
        return self.get_property(self.__INFO_URL_PROPERTY__)

    #########################           PROTECTED           #########################

    def _set_loaded(self, loaded, has_been_loaded=None):
        changed = False
        if self.loaded != loaded:
            self.loaded = loaded
            changed = True
        if has_been_loaded is None:
            if loaded:
                has_been_loaded = True
        if self._has_been_loaded != has_been_loaded:
            self._has_been_loaded = has_been_loaded
            changed = True
        if changed:
            self._set_dirty()

    def _set_property(self, name, value):
        if (not self.has_property(name)) or (self.get_property(name) != value):
            self._set_dirty()
        self.properties[name] = value

    def _reset_property(self, name, value):
        if self.has_property(name):
            self.properties.pop(name)
            self._set_dirty()



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
EMBLImageTracking
Hardware object used to control image tracking
By default ADXV is used
"""

from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLImageTracking(HardwareObject):
    """
    EMBLImageTracking
    """

    def __init__(self, *args):
        super().__init__(*args)

        self.state = None
        self.state_dict = {"image_tracking": False, "filter_frames": False}

        self.chan_state = None
        self.chan_enable_image_tracking = None
        self.chan_filter_frames = None
        self.chan_spot_list = None
        self.cmd_load_image = None

    def init(self):

        self.chan_enable_image_tracking = self.get_channel_object(
            "chanImageTrackingEnabled", optional=True
        )
        self.chan_enable_image_tracking.connect_signal(
            "update", self.image_tracking_state_changed
        )
        self.chan_filter_frames = self.get_channel_object(
            "chanFilterFramesEnabled", optional=True
        )
        if self.chan_filter_frames is not None:
            self.chan_filter_frames.connect_signal(
                "update", self.filter_frames_enabled_changed
            )

        self.chan_spot_list = self.get_channel_object(
            "chanSpotListEnabled", optional=True
        )
        if self.chan_spot_list is not None:
            self.chan_spot_list.connect_signal("update", self.spot_list_enabled_changed)

        self.chan_spot_list.set_value(True)
        self.chan_state = self.get_channel_object("chanState")
        self.chan_state.connect_signal("update", self.state_changed)

        self.cmd_load_image = self.get_command_object("cmdLoadImage")

    def image_tracking_state_changed(self, state):
        """
        Updates image tracking state
        :param state:
        :return:
        """
        self.state_dict["image_tracking"] = state
        self.emit("imageTrackingStateChanged", (self.state_dict,))

    def filter_frames_enabled_changed(self, state):
        """
        Updates filter frames state
        :param state:
        :return:
        """
        self.state_dict["filter_frames"] = state
        self.emit("imageTrackingStateChanged", (self.state_dict,))

    def state_changed(self, state):
        """
        Updates overall state
        :param state:
        :return:
        """
        if self.state != state:
            self.state = state
        self.emit("stateChanged", (self.state,))

    def is_tracking_enabled(self):
        """
        Returns True if image tracking is enabled
        :return:
        """
        return self.chan_enable_image_tracking.get_value()

    def set_image_tracking_state(self, state):
        """
        Enables/disables image tracking
        :param state:
        :return:
        """
        self.chan_enable_image_tracking.set_value(state)

    def set_filter_frames_state(self, state):
        """
        Enables/disables image filtering based on the dozor score
        :param state:
        :return:
        """
        self.chan_filter_frames.set_value(state)

    def set_spot_list_enabled(self, state):
        """
        Enables/disables spot indication on Adxv
        :param state:
        :return:
        """
        self.chan_spot_list.set_value(state)

    def spot_list_enabled_changed(self, state):
        self.state_dict["spot_list"] = state
        self.emit("imageTrackingStateChanged", (self.state_dict,))

    def load_image(self, image_name):
        """
        Load image in the image viewer
        :param image_name:
        :return:
        """
        if self.is_tracking_enabled():
            self.set_image_tracking_state(False)
        self.cmd_load_image(str(image_name))

    def re_emit_values(self):
        """
        Reemits values
        :return:
        """
        self.emit("stateChanged", self.state)
        self.emit("imageTrackingStateChanged", (self.state_dict,))



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import ast
import random

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.DataPublisher import (
    DataType,
    PlotDim,
    PlotType,
    one_d_data,
)


class ScanMockup(HardwareObject):
    """
    ScanMockup generates random data points to simulate an arbitrary scan
    """

    def __init__(self, name):
        super(ScanMockup, self).__init__(name)
        self._npoints = 100
        self._min = 0
        self._max = 10
        self._sample_rate = 0.5
        self._current_value = 0
        self._task = None
        self._dp_id = None

    def init(self):
        """
        FWK2 Init method
        """
        super(ScanMockup, self).init()

        self._npoints = self.get_property("number_of_points", 100)
        self._min, self._max = ast.literal_eval(self.get_property("min_max", "(0, 10)"))
        self._sample_rate = self.get_property("sample_rate", 0.5)

        HWR.beamline.data_publisher.register(
            "mockupscan",
            "MOCK SCAN",
            "diode",
            axis_labels=["time", "counts"],
            data_type=DataType.FLOAT,
            data_dim=PlotDim.ONE_D,
            plot_type=PlotType.SCATTER,
            content_type="MOCK SCAN",
            sample_rate=self._sample_rate,
            _range=(self._min, self._max),
            meta={},
        )

    def _generate_points(self):
        points = 0

        HWR.beamline.data_publisher.start("mockupscan")

        while points < self._npoints:
            self._current_value = random.uniform(self._min, self._max)

            HWR.beamline.data_publisher.pub(
                "mockupscan",
                one_d_data(points * self._sample_rate, self._current_value),
            )

            gevent.sleep(self._sample_rate)
            points += 1

        HWR.beamline.data_publisher.stop("mockupscan")

    def start(self):
        """
        Start scan
        """
        if not self._task:
            self._task = gevent.spawn(self._generate_points)

    def stop(self):
        """
        Stop scan
        """
        self._task.kill()
        self._task = None



"""
Example xml file
<object class="LNLSZoom">
  <username>zoom</username>
  <actuator_name>zoom</actuator_name>
  <exporter_address>130.235.94.124:9001</exporter_address>
  <values>{"LEVEL1": 1, "LEVEL2": 2, "LEVEL3": 3, "LEVEL4": 4, "LEVEL5": 5, "LEV
EL6": 6}</values>
</object>
"""
import time
from enum import Enum

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractNState import (
    AbstractNState,
    BaseValueEnum,
)
from mxcubecore.HardwareObjects.LNLS.EPICSActuator import EPICSActuator


class LNLSZoom(EPICSActuator, AbstractNState):
    """MicrodiffZoomMockup class"""

    def __init__(self, name):
        super(LNLSZoom, self).__init__(name)

    def init(self):
        """Initialize the zoom"""
        EPICSActuator.init(self)
        AbstractNState.init(self)

        self.initialise_values()
        _len = len(self.VALUES) - 1
        if _len > 0:
            # we can only assume that the values are consecutive integers
            # so the limits correspond to the keys
            limits = (1, _len)
            self.set_limits(limits)
        else:
            # Normally we get the limits from the hardware
            limits = (1, 10)
            self.set_limits(limits)
            # there is nothing in the xml file, create ValueEnum from the limits
            self._initialise_values()

        self.update_limits(limits)
        current_value = self.get_value()
        self.update_value(current_value)
        self.update_state(self.STATES.READY)

    def set_limits(self, limits=(None, None)):
        """Overrriden from AbstractActuator"""
        self._nominal_limits = limits

    def update_limits(self, limits=None):
        """Overrriden from AbstractNState"""
        if limits is None:
            limits = self.get_limits()

        self._nominal_limits = limits
        self.emit("limitsChanged", (limits,))

    def _initialise_values(self):
        """Initialise the ValueEnum """
        low, high = self.get_limits()

        values = {"LEVEL%s" % str(v): v for v in range(low, high + 1)}
        self.VALUES = Enum(
            "ValueEnum",
            dict(values, **{item.name: item.value for item in BaseValueEnum}),
        )

    def _move(self, value):
        """Override super class method."""
        self.update_state(self.STATES.BUSY)
        time.sleep(0.2)
        self.update_state(self.STATES.READY)
        current_value = self.get_value()
        self.update_value(current_value)
        return value

    def get_value(self):
        """Override super class method."""
        current_val = super(LNLSZoom, self).get_value()
        current_enum = self.value_to_enum(current_val)
        return current_enum

    def _set_value(self, value):
        """Override super class method."""
        enum = value
        target_val = enum.value
        super(LNLSZoom, self)._set_value(target_val)



import logging
import math
import sys
import time

from calc_flux import CalculateFlux

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import task


class ID29PhotonFlux(HardwareObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self):
        self.counter = self.get_object_by_role("counter")
        self.shutter = self.get_deviceby_role("shutter")
        self.aperture = self.get_object_by_role("aperture")
        fname = self.get_property("calibrated_diodes_file")

        self.flux_calc = CalculateFlux()
        self.flux_calc.init(fname)
        self.shutter.connect("shutterStateChanged", self.shutterStateChanged)

        self.counts_reading_task = self._read_counts_task(wait=False)

    @task
    def _read_counts_task(self):
        old_counts = None
        while True:
            counts = self._get_counts()
            if counts != old_counts:
                old_counts = counts
                self.countsUpdated(counts)
            time.sleep(1)

    def _get_counts(self):
        try:
            counts = self.counter.getCorrectedPhysValue()
            if counts == -9999:
                counts = 0
        except Exception:
            counts = 0
            logging.getLogger("HWR").exception("%s: could not get counts", self.name())
        try:
            egy = HWR.beamline.energy.get_value() * 1000.0
            calib = self.flux_calc.calc_flux_coef(egy)
        except Exception:
            logging.getLogger("HWR").exception("%s: could not get energy", self.name())
        else:
            if self.aperture is None:
                aperture_coef = 1
            else:
                try:
                    aperture_coef = self.aperture.getApertureCoef()
                except Exception:
                    sys.excepthook(*sys.exc_info())
                    aperture_coef = 1
            counts = math.fabs(counts * calib[0] * aperture_coef)
        return counts

    def connect_notify(self, signal):
        if signal == "valueChanged":
            self.emitValueChanged()

    def shutterStateChanged(self, _):
        self.countsUpdated(self._get_counts())

    def updateFlux(self, _):
        self.countsUpdated(self._get_counts(), ignore_shutter_state=False)

    def countsUpdated(self, counts, ignore_shutter_state=False):
        self.emitValueChanged("%1.3g" % counts)

    def get_value(self):
        self.updateFlux("dummy")
        return self.current_flux

    def emitValueChanged(self, flux=None):
        if flux is None:
            self.current_flux = None
            self.emit("valueChanged", ("?",))
        else:
            self.current_flux = flux
            self.emit("valueChanged", (self.current_flux,))



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""AbstractFlux class
Defines get_average_flux_density.
"""
from scipy.interpolate import interp1d

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator

__copyright__ = """ Copyright © 2010-2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractFlux(AbstractActuator):
    """Class for Flux abstraction"""

    def __init__(self, name):
        super().__init__(name)
        # flux by default is read only
        self.read_only = True

    def init(self):
        """Initialise some parameters."""
        super().init()
        self.read_only = self.get_property("read_only") or True

    # Dose rate for a standard composition crystal, in Gy/s
    # As a function of energy in keV
    #
    # NB this can be called as a function, the same as if it was defined thus:
    # def get_dose_rate_per_photon_per_mmsq(self, energy):
    #
    # The interpolation table get_dose_rate_per_photon_per_mmsq was created using
    # "Absorbed dose calculations for macromolecular crystals: improvements to RADDOSE"
    # Paithankar, K.S., Owen, R.L and Garman, E.F J. Syn. Rad. (2009), 16, 152-162,
    # for some sensible crystal composition, by Gleb Bourenkov
    #
    # The reason for *not* having a get_dose_rate function is that the actual dose rate
    # depends heavily on beam profile and crystal size and shape.
    # The necessary approximations should be done locally.
    # See GphlWorkflow for an example of how to do it.
    #
    get_dose_rate_per_photon_per_mmsq = interp1d(
        [4.0, 6.6, 9.2, 11.8, 14.4, 17.0, 19.6, 22.2, 24.8, 27.4, 30.0],
        [
            4590.0e-12,
            1620.0e-12,
            790.0e-12,
            457.0e-12,
            293.0e-12,
            202.0e-12,
            146.0e-12,
            111.0e-12,
            86.1e-12,
            68.7e-12,
            55.2e-12,
        ],
    )

    def get_average_flux_density(self, transmission=None):
        """Get average flux density over the beam area in photons / mm^2
        for a given transmission setting
        Args:
            transmission (float): Target transmission [%]
                                  (defaults to current value)
        Returns:
            (float): (photons / mm^2) - average flux density over beam area.
        """
        beam_size = HWR.beamline.beam.get_beam_size()
        flux = self.get_value()
        result = None
        if flux and all(beam_size):
            result = flux / (beam_size[0] * beam_size[1])
            if transmission is not None:
                result *= transmission / HWR.beamline.transmission.get_value()
        return result



from __future__ import print_function

import logging
import warnings

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractLims import AbstractLims
from mxcubecore.HardwareObjects.abstract.ISPyBDataAdapter import ISPyBDataAdapter
from mxcubecore.model.lims_session import LimsSessionManager


class ISPyBAbstractLIMS(AbstractLims):
    """
    Web-service client for ISPyB.
    """

    def __init__(self, name):
        super().__init__(name)
        self.ldapConnection = None
        self.pyispyb = None
        self._translations = {}
        self.authServerType = None
        self.loginTranslate = None
        self.base_result_url = None
        self.login_ok = False

    def init(self):
        super().init()
        self.pyispyb = self.get_object_by_role("pyispyb")
        self.icat_client = self.get_object_by_role("icat_client")

        self.samples = []
        self.authServerType = self.get_property("authServerType") or "ldap"
        if self.authServerType == "ldap":
            # Initialize ldap
            self.ldapConnection = self.get_object_by_role("ldapServer")
            if self.ldapConnection is None:
                logging.getLogger("HWR").debug("LDAP Server is not available")

        self.loginTranslate = self.get_property("loginTranslate") or True

        # ISPyB Credentials
        self.ws_root = self.get_property("ws_root")
        self.ws_username = self.get_property("ws_username") or None
        self.ws_password = str(self.get_property("ws_password")) or None

        self.proxy_address = self.get_property("proxy_address")
        if self.proxy_address:
            self.proxy = {"http": self.proxy_address, "https": self.proxy_address}
        else:
            self.proxy = {}

        try:
            self.base_result_url = self.get_property("base_result_url").strip()
        except AttributeError:
            pass

        self.adapter = ISPyBDataAdapter(
            self.ws_root.strip(),
            self.proxy,
            self.ws_username,
            self.ws_password,
            self.beamline_name,
        )
        logging.getLogger("HWR").debug("[ISPYB] Proxy address: %s" % self.proxy)

        # Add the porposal codes defined in the configuration xml file
        # to a directory. Used by translate()
        if hasattr(HWR.beamline.session, "proposals"):
            for proposal in HWR.beamline.session["proposals"]:
                code = proposal.code
                self._translations[code] = {}
                try:
                    self._translations[code]["ldap"] = proposal.ldap
                except AttributeError:
                    pass
                try:
                    self._translations[code]["ispyb"] = proposal.ispyb
                except AttributeError:
                    pass
                try:
                    self._translations[code]["gui"] = proposal.gui
                except AttributeError:
                    pass

    def get_user_name(self):
        raise Exception("Abstract class. Not implemented")

    def get_full_user_name(self):
        raise Exception("Abstract class. Not implemented")

    def is_user_login_type(self):
        raise Exception("Abstract class. Not implemented")

    def store_beamline_setup(self, session_id, bl_config):
        self.adapter.store_beamline_setup(session_id, bl_config)

    def _translate(self, code, what):
        try:
            translated = self._translations[code][what]
        except KeyError:
            translated = code
        return translated

    def echo(self):
        if not self.adapter._shipping:
            msg = "Error in echo: Could not connect to server."
            logging.getLogger("ispyb_client").warning(msg)
            raise Exception("Error in echo: Could not connect to server.")

        try:
            self.adapter._shipping._shipping.service.echo()
            return True
        except Exception as e:
            logging.getLogger("ispyb_client").error(str(e))

        return False

    def ldap_login(self, login_name, psd):
        warnings.warn(
            (
                "Using Authenticatior from ISPyBClient is deprecated,"
                "use Authenticator to authenticate spereatly and then login to ISPyB"
            ),
            DeprecationWarning,
        )

        return self.ldapConnection.authenticate(login_name, psd)

    def store_data_collection(self, mx_collection, bl_config=None):
        return self._store_data_collection(mx_collection, bl_config)

    def update_data_collection(self, mx_collection):
        return self._update_data_collection(mx_collection)

    def finalize_data_collection(self, mx_collection):
        # Also upload the same data to icat if icat_client is available
        if self.icat_client:
            self.icat_client.store_data_collection(mx_collection)

        return self._update_data_collection(mx_collection)

    def _store_data_collection(self, mx_collection, bl_config=None):
        return self.adapter.store_data_collection(mx_collection, bl_config)

    def _update_data_collection(self, mx_collection):
        return self.adapter._update_data_collection(mx_collection)

    def update_bl_sample(self, bl_sample):
        return self.adapter.update_bl_sample(bl_sample)

    def store_image(self, image_dict):
        self.adapter.store_image(image_dict)

    def find_sample_by_sample_id(self, sample_id):
        for sample in self.samples:
            try:
                if int(sample.get("limsID")) == sample_id:
                    return sample
            except (TypeError, KeyError):
                pass
        return None

    def get_samples(self, lims_name):
        self.samples = []
        if self.session_manager.active_session is not None:
            self.samples = self.adapter.get_samples(
                self.session_manager.active_session.proposal_id
            )
            logging.getLogger("HWR").debug(
                "get_samples. %s samples retrieved. proposal_id=%s lims_name=%s"
                % (
                    len(self.samples),
                    self.session_manager.active_session.proposal_id,
                    lims_name,
                )
            )
        return self.samples

    def create_session(self, proposal_id: str):
        logging.getLogger("HWR").debug("create_session. proposal_id=%s" % proposal_id)
        session_manager = self.adapter.create_session(proposal_id, self.beamline_name)
        logging.getLogger("HWR").debug("Session created. proposal_id=%s" % proposal_id)
        return session_manager

    def store_energy_scan(self, energyscan_dict):
        return self.adapter.store_energy_scan(energyscan_dict)

    def associate_bl_sample_and_energy_scan(self, entry_dict):
        return self.adapter.associate_bl_sample_and_energy_scan(entry_dict)

    def get_data_collection(self, data_collection_id):
        return self.adapter.get_data_collection(self, data_collection_id)

    def get_session(self, session_id):
        return self.adapter.get_session(session_id)

    def store_xfe_spectrum(self, xfespectrum_dict):
        return self.adapter.store_xfe_spectrum(xfespectrum_dict)

    def is_connected(self):
        return self.login_ok

    def isInhouseUser(self, proposal_code, proposal_number):
        for proposal in self["inhouse"]:
            if proposal_code == proposal.code:
                if str(proposal_number) == str(proposal.number):
                    return True
        return False

    def _store_data_collection_group(self, group_data):
        return self.adapter._store_data_collection_group(group_data)

    def store_workflow(self, *args, **kwargs):
        try:
            return self._store_workflow(*args, **kwargs)
        except gevent.GreenletExit:
            raise
        except Exception:
            logging.exception("Could not store workflow")
            return None, None, None

    def store_robot_action(self, robot_action_dict):
        return self.adapter.store_robot_action(robot_action_dict)

    def create_mx_collection(self, collection_parameters):
        self.icat_client.create_mx_collection(collection_parameters)

    def create_ssx_collection(
        self, data_path, collection_parameters, beamline_parameters, extra_lims_values
    ):
        self.icat_client.create_ssx_collection(
            data_path, collection_parameters, beamline_parameters, extra_lims_values
        )



"""
XMLRPC-Server that makes it possbile to access core features of MXCuBE like
the queue from external applications. The Server is implemented as a
hardware object and is configured with an XML-file. See the example
configuration XML for more information.
"""

import atexit
import inspect
import json
import logging
import os
import pkgutil
import shutil
import socket
import sys
import time
import types
import xml
from functools import reduce

import gevent
import jsonpickle

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.SecureXMLRpcRequestHandler import (
    SecureXMLRpcRequestHandler,
)

if sys.version_info > (3, 0):
    from xmlrpc.server import SimpleXMLRPCServer
else:
    from SimpleXMLRPCServer import SimpleXMLRPCServer


__author__ = "Marcus Oskarsson, Matias Guijarro"
__copyright__ = "Copyright 2012, ESRF"
__credits__ = ["MxCuBE collaboration"]

__version__ = ""
__maintainer__ = "Marcus Oskarsson"
__email__ = "marcus.oscarsson@esrf.fr"
__status__ = "Draft"


class XMLRPCServer(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.host = None
        self.port = None
        self.all_interfaces = None
        self.enforceUseOfToken = None

        self.wokflow_in_progress = True
        self.xmlrpc_prefixes = set()
        self.current_entry_task = None
        self.host = None
        self.use_token = None

        atexit.register(self.close)
        self.gphl_workflow_status = None

    def init(self):
        """
        Method inherited from HardwareObject, called by framework-2.
        """
        self.all_interfaces = self.get_property("all_interfaces", False)
        # Listen on all interfaces if <all_interfaces>True</all_interfaces>
        # otherwise only on the interface corresponding to socket.gethostname()
        if self.all_interfaces:
            host = ""
        else:
            host = socket.gethostname()

        self.host = host
        self.port = self.get_property("port")

        self.use_token = self.get_property("use_token", False)

        try:
            self.open()
        except Exception:
            logging.getLogger("HWR").debug("Can't start XML-RPC server")

    def close(self):
        try:
            self.xmlrpc_server_task.kill()
            self._server.server_close()
            del self._server
        except AttributeError:
            pass

    def open(self):
        # The value of the member self.port is set in the xml configuration
        # file. The initialization is done by the baseclass HardwareObject.
        if hasattr(self, "_server"):
            return
        self.xmlrpc_prefixes = set()

        if self.use_token:
            self._server = SimpleXMLRPCServer(
                (self.host, int(self.port)),
                requestHandler=SecureXMLRpcRequestHandler,
                logRequests=False,
                allow_none=True,
            )
        else:
            self._server = SimpleXMLRPCServer(
                (self.host, int(self.port)), logRequests=False, allow_none=True
            )

        msg = "XML-RPC server listening on: %s:%s" % (self.host, self.port)
        logging.getLogger("HWR").info(msg)

        self.connect(
            HWR.beamline.gphl_workflow,
            "gphl_workflow_finished",
            self._async_job_completed,
        )

        self._server.register_introspection_functions()
        self._server.register_function(self.start_queue)
        self._server.register_function(self.log_message)
        self._server.register_function(self.is_queue_executing)
        self._server.register_function(self.queue_execute_entry_with_id)
        self._server.register_function(self.queue_set_workflow_lims_id)
        self._server.register_function(self.shape_history_get_grid)
        self._server.register_function(self.shape_history_set_grid_data)
        self._server.register_function(self.beamline_setup_read)
        self._server.register_function(self.get_default_path_template)
        self._server.register_function(self.get_default_acquisition_parameters)

        self._server.register_function(self.get_diffractometer_positions)
        self._server.register_function(self.get_resolution_limits)
        self._server.register_function(self.move_diffractometer)
        self._server.register_function(self.save_snapshot)
        self._server.register_function(self.save_multiple_snapshots)
        self._server.register_function(self.save_twelve_snapshots_script)
        self._server.register_function(self.cryo_temperature)
        self._server.register_function(self.flux)
        self._server.register_function(self.check_for_beam)
        self._server.register_function(self.set_beam_size)
        self._server.register_function(self.get_beam_size)
        self._server.register_function(self.get_available_beam_size)
        self._server.register_function(self.set_aperture)
        self._server.register_function(self.get_aperture)
        self._server.register_function(self.get_aperture_list)
        self._server.register_function(self.get_cp)
        self._server.register_function(self.save_current_pos)
        self._server.register_function(self.anneal)
        self._server.register_function(self.open_dialog)
        self._server.register_function(self.workflow_end)
        self._server.register_function(self.dozor_batch_processed)
        self._server.register_function(self.dozor_status_changed)
        self._server.register_function(self.processing_status_changed)
        self.image_num = 0
        self._server.register_function(self.get_image_num, "get_image_num")
        self._server.register_function(self.set_zoom_level)
        self._server.register_function(self.get_zoom_level)
        self._server.register_function(self.get_available_zoom_levels)
        self._server.register_function(self.set_front_light_level)
        self._server.register_function(self.get_front_light_level)
        self._server.register_function(self.set_back_light_level)
        self._server.register_function(self.get_back_light_level)
        self._server.register_function(self.centre_beam)
        self._server.register_function(self.addXrayCentring)
        self._server.register_function(self.addGphlWorkflow)
        self._server.register_function(self.get_gphl_workflow_status)
        self._server.register_function(self.clearISPyBClientGroupId)
        self._server.register_function(self.setCharacterisationResult)

        # Register functions from modules specified in <apis> element
        if self.has_object("apis"):
            apis = next(self.get_objects("apis"))
            for api in apis.get_objects("api"):
                recurse = api.get_property("recurse")
                if recurse is None:
                    recurse = True

                self._register_module_functions(
                    api.get_property("module"), recurse=recurse
                )

        self.xmlrpc_server_task = gevent.spawn(self._server.serve_forever)
        self.beamcmds_hwobj = self.get_object_by_role("beamcmds")

    def anneal(self, time):
        cryoshutter_hwobj = self.get_object_by_role("cryoshutter")
        try:
            cryoshutter_hwobj.getCommandObject("anneal")(time)
        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise
        return True

    def _add_to_queue(self, task, set_on=True):
        """
        Adds the TaskNode objects contained in the
        list of TaskNodes passed in <task>.

        The TaskNodes are marked as activated in the queue if <set_on>
        is True and to inactivated if False.

        :param task: TaskNode object to add to queue
        :type parent: TaskNode

        :param set_on: Mark TaskNode as activated if True and as inactivated
                       if false.
        :type set_on: bool

        :returns: True on success otherwise False
        :rtype: bool
        """

        # The exception is re raised so that it will
        # be sent to the client.
        try:
            self.emit("add_to_queue", (task, None, set_on))

        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise
        return True

    def start_queue(self):
        """
        Starts the queue execution.

        :returns: True on success otherwise False
        :rtype: bool
        """
        try:
            self.emit("start_queue")
        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise
        return True

    def log_message(self, message, level="info"):
        """
        Logs a message in the user_level_log of MxCuBE,
        normally displayed at the bottom of the MxCuBE
        window.

        :param message: The message to log
        :type parent: str

        :param message: The log level, one of the strings:
                        'info'. 'warning', 'error'
        :type parent: str

        :returns: True on success otherwise False
        :rtype: bool
        """
        status = True

        if level == "info":
            logging.getLogger("user_level_log").info(message)
        elif level == "warning":
            logging.getLogger("user_level_log").warning(message)
        elif level == "error":
            logging.getLogger("user_level_log").error(message)
        else:
            status = False

        return status

    def _model_add_child(self, parent_id, child):
        """
        Adds the model node task to parent_id.

        :param parent_id: The id of the parent.
        :type parent_id: int

        :param child: The TaskNode object to add.
        :type child: TaskNode

        :returns: The id of the added TaskNode object.
        :rtype: int
        """
        try:
            node_id = HWR.beamline.queue_model.add_child_at_id(parent_id, child)
        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise
        return node_id

    def _model_get_node(self, node_id):
        """
        :returns the TaskNode object with the node id <node_id>
        :rtype: TaskNode
        """
        try:
            node = HWR.beamline.queue_model.get_node(node_id)
        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise
        return node

    def queue_execute_entry_with_id(self, node_id, use_async=False):
        """
        Execute the entry that has the model with node id <node_id>.

        :param node_id: The node id of the model to find.
        :type node_id: int
        """
        try:
            model = HWR.beamline.queue_model.get_node(node_id)
            entry = HWR.beamline.queue_manager.get_entry_with_model(model)

            if entry:
                self.current_entry_task = HWR.beamline.queue_manager.execute_entry(
                    entry, use_async=use_async
                )

        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise
        return True

    def queue_set_workflow_lims_id(self, node_id, lims_id):
        """
        Set lims id of workflow node with id <node_id>

        :param node_id: The node id of the workflow node
        :type node_id: int
        :param lims_id: The lims id
        :type lims_id: int
        """
        try:
            model = HWR.beamline.queue_model.get_node(node_id)
            model.lims_id = lims_id
        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise
        else:
            return True

    def is_queue_executing(self, node_id=None):
        """
        :returns: True if the queue is executing otherwise False
        :rtype: bool
        """
        try:
            return HWR.beamline.queue_manager.is_executing(node_id)
        except Exception as ex:
            logging.getLogger("HWR").exception(str(ex))
            raise

    def queue_status(self):
        pass

    def shape_history_get_grid(self, sid):
        """
        :param sid: Shape id
        :returns: Grid with id <sid>
        :rtype: dict

        Format of the returned dictionary:

        {'id': id,
         'dx_mm': float,
         'dy_mm': float,
         'steps_x': int,
         'steps_y': int,
         'x1': float,
         'y1': float,
         'angle': float}

        """
        grid_dict = HWR.beamline.sample_view.get_shape(sid).as_dict()

        return grid_dict

    def shape_history_set_grid_data(self, key, result_data, data_file_path=None):
        if isinstance(result_data, list):
            result = {}

            for result in result_data.items():
                # int_based_result is not defined
                int_based_result[int(result[0])] = result[1]
        else:
            result = result_data

        HWR.beamline.sample_view.set_grid_data(key, result, data_file_path)
        return True

    def get_cp(self):
        """
        :returns: a json encoded list with all centred positions
        """
        cplist = []
        points = HWR.beamline.sample_view.get_points()

        for point in points:
            cp = point.get_centred_positions()[0].as_dict()
            cplist.append(cp)

        json_cplist = json.dumps(cplist)

        return json_cplist

    def _getattr_from_path(self, obj, attr, delim="/"):
        """Recurses through an attribute chain to get the attribute."""
        return reduce(getattr, attr.split(delim), obj)

    def beamline_setup_read(self, path):
        value = None

        if path.strip("/").endswith("default-acquisition-parameters"):
            value = jsonpickle.encode(self.get_default_acquisition_parameters())
        elif path.strip("/").endswith("default-path-template"):
            value = jsonpickle.encode(self.get_default_path_template())
        else:
            try:
                path = path[1:] if path[0] == "/" else path
                ho = self._getattr_from_path(HWR, path)
                value = ho.get_value()
            except:
                logging.getLogger("HWR").exception("Could no get %s " % str(path))

        return value

    def get_default_path_template(self):
        return HWR.beamline.get_default_path_template()

    def get_default_acquisition_parameters(self):
        return HWR.beamline.get_default_acquisition_parameters()

    def workflow_set_in_progress(self, state):
        if state:
            self.wokflow_in_progress = True
        else:
            self.wokflow_in_progress = False

    def get_resolution_limits(self):
        return HWR.beamline.resolution.get_limits()

    def get_diffractometer_positions(self):
        return HWR.beamline.diffractometer.get_positions()

    def move_diffractometer(self, roles_positions_dict):
        HWR.beamline.diffractometer.move_motors(roles_positions_dict)
        return True

    def save_twelve_snapshots_script(self, path):
        HWR.beamline.diffractometer.run_script("Take12Snapshots")
        # Wait a couple of seconds for the files to appear

        time.sleep(2)
        HWR.beamline.diffractometer.wait_ready(300)
        tmp_path = HWR.beamline.diffractometer.get_property(
            "custom_snapshot_script_dir", "/tmp"
        )

        file_list = os.listdir(tmp_path)

        for filename in file_list:
            shutil.copy(tmp_path + filename, path)

    def save_multiple_snapshots(self, path_list, show_scale=False):
        logging.getLogger("HWR").info("Taking snapshot %s " % str(path_list))

        try:
            for angle, path in path_list:
                HWR.beamline.diffractometer.phiMotor.set_value(angle)
                # give some time to get the snapshot
                time.sleep(1)
                HWR.beamline.diffractometer.wait_ready()
                self.save_snapshot(path, show_scale, handle_light=False)
        except Exception as ex:
            logging.getLogger("HWR").exception("Could not take snapshot %s " % str(ex))

    def save_snapshot(self, imgpath, showScale=False, handle_light=True):
        res = True
        logging.getLogger("HWR").info("Taking snapshot %s " % str(imgpath))

        try:
            if showScale:
                HWR.beamline.diffractometer.save_snapshot(imgpath)
            else:
                HWR.beamline.sample_view.save_snapshot(imgpath, overlay=False, bw=False)
        except Exception as ex:
            logging.getLogger("HWR").exception("Could not take snapshot %s " % str(ex))
            res = False
        finally:
            pass

        return res

    def save_current_pos(self):
        """
        Saves the current position as a centered position.
        """
        HWR.beamline.diffractometer.save_current_position()
        return True

    def cryo_temperature(self):
        return HWR.beamline.diffractometer.cryostream.get_value()

    def flux(self):
        flux = HWR.beamline.flux.get_value()
        if flux is None:
            flux = 0
        return float(flux)

    def check_for_beam(self):
        return HWR.beamline.flux.is_beam()

    def set_beam_size(self, size):
        """Set the beam size.
        Args:
            size (list): Width, heigth or
                 (str): Size label.
        """
        HWR.beamline.beam.set_value(size)
        return True

    def get_beam_size(self):
        """Get the beam size [um], its shape and label.
        Returns:
            (tuple):  (width, heigth, shape, label), with types
                      (float, float, str, str)
        """
        return HWR.beamline.beam.get_value_xml()

    def get_available_beam_size(self):
        """Get the available predefined beam sizes.
        Returns:
            (dict): Dictionary wiith list of avaiable beam size labels
                    and the corresponding size (width,height) tuples.
                    {"label": [str, str, ...], "size": [(w,h), (w,h), ...]}
        """
        return HWR.beamline.beam.get_defined_beam_size()

    def set_aperture(self, pos_name):
        HWR.beamline.beam.set_value(pos_name)
        return True

    def get_aperture(self):
        return HWR.beamline.beam.get_value()[-1]

    def get_aperture_list(self):
        return HWR.beamline.beam.get_available_size()["values"]

    def open_dialog(self, dict_dialog):
        """
        Opens the workflow dialog in mxCuBE.
        This call blocks util the dialog is ended by the user.
        """

        return_map = {}
        workflow_hwobj = HWR.beamline.workflow
        if workflow_hwobj is not None:
            return_map = workflow_hwobj.open_dialog(dict_dialog)
        self.emit("open_dialog", dict_dialog)
        return return_map

    def workflow_end(self):
        """
        Notify the workflow HO that the workflow has finished.
        """
        workflow_hwobj = HWR.beamline.workflow
        if workflow_hwobj is not None:
            workflow_hwobj.workflow_end()

    def dozor_batch_processed(self, dozor_batch_dict):
        HWR.beamline.online_processing.batch_processed(dozor_batch_dict)

    def dozor_status_changed(self, status):
        HWR.beamline.online_processing.set_processing_status(status)

    def processing_status_changed(self, collection_id, method, status, msg=""):
        for queue_entry in HWR.beamline.queue_model.get_all_dc_queue_entries():
            data_model = queue_entry.get_data_model()
            if data_model.id == collection_id:
                prefix = data_model.acquisitions[0].path_template.get_image_file_name()
                prefix = prefix.replace("%05d", "#####")

                if status in ("started", "success"):
                    logging.getLogger("user_level_log").info(
                        "EDNA %s: processing of data collection %s %s %s"
                        % (method, prefix, status, msg)
                    )
                elif status == "failed":
                    logging.getLogger("user_level_log").error(
                        "EDNA %s: processing of data collection %s %s %s"
                        % (method, prefix, status, msg)
                    )

                queue_entry.add_processing_msg(
                    str(time.strftime("%Y-%m-%d %H:%M:%S")), method, status, msg
                )

    def image_taken(self, image_num):
        self.image_num = image_num

    def get_image_num(self):
        return self.image_num

    def set_zoom_level(self, pos):
        """
        Sets the zoom to a pre-defined level.
        """
        zoom = HWR.beamline.diffractometer.zoomMotor
        zoom.set_value(zoom.value_to_enum(pos))

    def get_zoom_level(self):
        """
        Returns the zoom level.
        """
        zoom = HWR.beamline.diffractometer.zoomMotor
        pos = zoom.get_value().value
        return pos

    def get_available_zoom_levels(self):
        """
        Returns the avaliable pre-defined zoom levels.
        """
        _value_enum = HWR.beamline.diffractometer.zoomMotor.VALUES.items()
        _names = [name for name, value in _value_enum.items()]

        return _names

    def set_front_light_level(self, level):
        """
        Sets the level of the front light
        """
        HWR.beamline.diffractometer.setFrontLightLevel(level)

    def get_front_light_level(self):
        """
        Gets the level of the front light
        """
        return HWR.beamline.diffractometer.getFrontLightLevel()

    def set_back_light_level(self, level):
        """
        Sets the level of the back light
        """
        logging.getLogger("HWR").info("Setting backlight level to %s" % level)
        HWR.beamline.diffractometer.setBackLightLevel(level)

    def get_back_light_level(self):
        """
        Gets the level of the back light
        """
        return HWR.beamline.diffractometer.getBackLightLevel()

    def centre_beam(self):
        """
        Centers the beam using the beamcmds hardware object.
        """
        self.beamcmds_hwobj.centrebeam()
        while (
            self.beamcmds_hwobj.centrebeam._cmd_execution
            and not self.beamcmds_hwobj.centrebeam._cmd_execution.ready()
        ):
            time.sleep(1)

    def _register_module_functions(self, module_name, recurse=True, prefix=""):
        log = logging.getLogger("HWR")
        # log.info("Registering functions in module %s with XML-RPC server" % module_name)

        if module_name not in sys.modules:
            __import__(module_name)
        module = sys.modules[module_name]

        if not hasattr(module, "xmlrpc_prefix"):
            log.error(
                (
                    'Module %s  has no attribute "xmlrpc_prefix": cannot '
                    + "register its functions. Skipping"
                )
                % module_name
            )
        else:
            prefix += module.xmlrpc_prefix
            if len(prefix) > 0 and prefix[-1] != "_":
                prefix += "_"

            if prefix in self.xmlrpc_prefixes:
                msg = "Prefix %s already used: cannot register for module %s" % (
                    prefix,
                    module_name,
                )
                log.error(msg)
                raise Exception(msg)
            self.xmlrpc_prefixes.add(prefix)

            for f in inspect.getmembers(module, inspect.isfunction):
                if f[0][0] != "_":
                    xmlrpc_name = prefix + f[0]
                    # log.info(
                    #    "Registering function %s.%s as XML-RPC function %s"
                    #    % (module_name, f[1].__name__, xmlrpc_name)
                    # )

                    # Bind method to this XMLRPCServer instance but don't set attribute
                    # This is sufficient to register it as an xmlrpc function.
                    bound_method = types.MethodType(f[1], self)
                    self._server.register_function(bound_method, xmlrpc_name)

            # TODO: Still need to test with deeply-nested modules, in particular that
            # modules and packages are both handled correctly in complex cases.
            if recurse and hasattr(module, "__path__"):
                sub_modules = pkgutil.walk_packages(module.__path__)
                try:
                    sub_module = next(sub_modules)
                    self._register_module_functions(
                        module_name + "." + sub_module[1], recurse=False, prefix=prefix
                    )
                except StopIteration:
                    pass

    def setToken(self, token):
        SecureXMLRpcRequestHandler.setReferenceToken(token)

    def clearISPyBClientGroupId(self):
        HWR.beamline.lims.group_id = None

    def setCharacterisationResult(self, characterisationResult):
        HWR.beamline.characterisation.characterisationResult = (
            xml.sax.saxutils.unescape(characterisationResult)
        )

    def addXrayCentring(self, parent_node_id, **centring_parameters):
        """Add Xray centring to queue."""
        from mxcubecore.model import queue_model_objects as qmo

        xc_model = qmo.XrayCentring2(**centring_parameters)
        child_id = HWR.beamline.queue_model.add_child_at_id(parent_node_id, xc_model)
        return child_id

    def addGphlWorkflow(self, parent_node_id, task_dict, workflow_id):
        """Add GPhL workflow to queue."""
        self.workflow_id = workflow_id
        from mxcubecore.model import queue_model_objects as qmo

        gphl_model = qmo.GphlWorkflow()
        parent_model = HWR.beamline.queue_model.get_node(int(parent_node_id))
        sample_model = parent_model.get_sample_node()
        gphl_model.init_from_task_data(sample_model, task_dict)
        child_id = HWR.beamline.queue_model.add_child_at_id(parent_node_id, gphl_model)
        self.gphl_workflow_status = "RUNNING"
        return child_id

    def _async_job_completed(self, job_status):
        self.gphl_workflow_status = job_status

    def get_gphl_workflow_status(self):
        return self.gphl_workflow_status



import logging

from ESRFEnergyScan import (
    ESRFEnergyScan,
    TunableEnergy,
)

from mxcubecore.TaskUtils import task


class BM14EnergyScan(ESRFEnergyScan):
    def __init__(self, name):
        ESRFEnergyScan.__init__(self, name, TunableEnergy())

    @task
    def energy_scan_hook(self, energy_scan_parameters):
        self.energy = energy_scan_parameters["edgeEnergy"]
        if self.energy_scan_parameters["findattEnergy"]:
            ESRFEnergyScan.move_energy(self, energy_scan_parameters["findattEnergy"])

    @task
    def move_undulators(self, gaps):
        return

    @task
    def set_mca_roi(self, eroi_min, eroi_max):
        self.execute_command("calculateMcaRoi", eroi_min, eroi_max)

    @task
    def choose_attenuation(self):
        if self.execute_command("chooseAttenuation") == -1:
            logging.getLogger("user_level_log").error(
                "Cannot find appropriate attenuation"
            )
            raise RuntimeError("Cannot find appropriate attenuation")
        self.energy_scan_parameters["transmissionFactor"] = (
            self.transmission.get_value()
        )

    @task
    def execute_energy_scan(self, energy_scan_parameters):
        energy_scan_parameters["exposureTime"] = self.get_property("exposureTime")
        self.execute_command("executeScan", energy_scan_parameters)
        self.energy_scan_parameters["exposureTime"] = energy_scan_parameters[
            "exposureTime"
        ]



import logging
import os.path

import numpy

from mxcubecore.HardwareObjects.XRFSpectrum import XRFSpectrum

try:
    from PyMca import (
        ClassMcaTheory,
        ConfigDict,
        QtMcaAdvancedFitReport,
    )
except ImportError:
    from PyMca5.PyMca import (
        ClassMcaTheory,
        ConfigDict,
        QtMcaAdvancedFitReport,
    )

"""
Next two lines is a trick to avoid core dump in QtMcaAdvancedFitReport
"""
from unittest.mock import MagicMock

QtMcaAdvancedFitReport.qt = MagicMock()


class ID29XRFSpectrum(XRFSpectrum):
    def __init__(self, *args, **kwargs):
        XRFSpectrum.__init__(self, *args, **kwargs)
        self.mca_hwobj = self.get_object_by_role("mca")
        self.ctrl_hwobj = self.get_object_by_role("controller")
        self.beamsize = self.get_object_by_role("beamsize")
        self.fname = self.get_property("cfgfile")
        self.config = ConfigDict.ConfigDict()
        self.mcafit = ClassMcaTheory.McaTheory(self.fname)
        self.default_integration_time = self.get_property("default_integration_time", 3.5)

    def preset_mca(self, ctime=None, fname=None):
        try:
            ctime = float(ctime)
        except (TypeError, ValueError):
            ctime = self.default_integration_time
    
        self.mca_hwobj.set_roi(2, 15, channel=1)
        self.mca_hwobj.set_presets(erange=1, ctime=ctime, fname=str(fname))
        self.ctrl_hwobj.mca.set_roi(2, 15, channel=1)
        self.ctrl_hwobj.mca.set_presets(erange=1, ctime=ctime, fname=str(fname))

    def _doSpectrum(self, ctime, filename, wait=True):
        try:
            ctime = float(ctime)
        except (TypeError, ValueError):
            ctime = self.default_integration_time
        print(f"_doSpectrum ctime {ctime}")
        self.choose_attenuation(ctime, filename)

    def choose_attenuation(self, ctime=None, fname=None):
        """Choose appropriate maximum attenuation.
        Args:
            ctime (float): count time [s]
        Kwargs:
            fname (str): Filename to save the MCA data (full path)
        Returns:
            (bool): Procedure executed correcly (True) or error (False)
        """
        try:
            ctime = float(ctime)
        except (TypeError, ValueError):
            ctime = self.default_integration_time
        print(f"ctime {ctime}")

        res = True
        if not fname:
            fname = self.spectrumInfo["filename"]
        fname = str(fname)

        self.preset_mca(ctime, fname)

        self.ctrl_hwobj.detcover.set_in(20)
        try:
            _transm = self.ctrl_hwobj.find_max_attenuation(
                ctime=ctime, fname=fname, roi=[2.0, 15.0]
            )
            self.spectrumInfo["beamTransmission"] = _transm
        except Exception as exp:
            logging.getLogger("user_level_log").exception(str(exp))
            res = False

        self.ctrl_hwobj.diffractometer.fldet_out()
        return res

    def _findAttenuation(self, ctime=None):
        try:
            ctime = float(ctime)
        except (TypeError, ValueError):
            ctime = self.default_integration_time
        return self.choose_attenuation(ctime)

    """
    Next methods are for fitting the data with pymca
    """

    def mcafit_configuration(self, config=None):
        """Configure the fitting parameters. The procedure is time consuming.
           It is only executed if the last configuration file is not the same.
        Args:
            config(dict): Configuration dictionary, containing among others the
                          configuration file name.
        """
        change = False
        if not config or "file" not in config:
            fname = XRFSpectrum._get_cfgfile(self, self.spectrumInfo["energy"])
        else:
            fname = config["file"]

        if self.fname != fname:
            self.fname = fname
            change = True
        self.config.read(self.fname)
        if "concentrations" not in self.config:
            self.config["concentrations"] = {}
            change = True
        if "attenuators" not in self.config:
            self.config["attenuators"] = {"Matrix": [1, "Water", 1.0, 0.01, 45.0, 45.0]}
            change = True
        if "flux" in config:
            self.config["concentrations"]["flux"] = float(config["flux"])
            change = True
        if "time" in config:
            self.config["concentrations"]["time"] = float(config["time"])
            change = True

        if change:
            self.mcafit.configure(self.config)

    def set_data(self, data, calib=None, config=None):
        """Execute the fitting. Write the fitted data files to pyarch.
        Args:
            data (list): The raw data.
            calib (list): The mca calibration.
            config (dict): The configuration dictionary.
        """
        if config:
            self.mcafit_configuration(config)
        try:
            if data[0].size == 2:
                xdata = numpy.array(data[:, 0]) * 1.0
                ydata = numpy.array(data[:, 1])
            else:
                xdata = data[0] * 1.0
                ydata = data[1]

            # xmin and xmax hard coded while waiting for configuration file
            # to be corrected.
            # xmin = 292
            # xmax = 4000

            xmin = int(config["min"])
            xmax = int(config["max"])
            self.mcafit.setData(xdata, ydata, xmin=xmin, xmax=xmax, calibration=calib)

            self.mcafit.estimate()
            # fitresult  = self._fit()

            fitresult = self.mcafit.startfit(digest=1)
            if fitresult:
                fitresult = {"fitresult": fitresult[0], "result": fitresult[1]}

                # write the csv file to pyarch
                csvname = self.spectrumInfo["fittedDataFileFullPath"]
                self._write_csv_file(fitresult, csvname)

                # write html report to pyarch
                fname = os.path.basename(self.spectrumInfo["filename"])
                outfile = fname.split(".")[0]
                outdir = os.path.dirname(self.spectrumInfo["annotatedPymcaXfeSpectrum"])

                _kw = {
                    "outdir": outdir,
                    "outfile": outfile,
                    "fitresult": fitresult,
                    "plotdict": {"logy": False},
                }

                report = QtMcaAdvancedFitReport.QtMcaAdvancedFitReport(**_kw)
                text = report.getText()
                report.writeReport(text=text)
        except Exception as exp:
            logging.getLogger().exception("XRFSpectrum: problem fitting %s" % str(exp))
            raise

    def _write_csv_file(self, fitresult, fname=None):
        """Write data to a csv file.
        Args:
            fitresult(dict): Data as dictionary.
        Kwargs:
            fname (str): Filename to write to (full path).
        """
        if not fname:
            fname = self.spectrumInfo["fittedDataFileFullPath"]
        if os.path.exists(fname):
            os.remove(fname)

        # get the significant peaks
        peaks_dict = {}
        pars_len = len(fitresult["result"]["parameters"])
        grp_len = len(fitresult["result"]["groups"])
        nglobal = pars_len - grp_len
        parameters = fitresult["result"]["fittedpar"][:nglobal] + [0.0] * grp_len

        for grp in fitresult["result"]["parameters"][nglobal:]:
            idx = fitresult["result"]["parameters"].index(grp)
            parameters[idx] = fitresult["result"]["fittedpar"][idx]
            xmatrix = fitresult["result"]["xdata"]
            ymatrix = self.mcafit.mcatheory(parameters, xmatrix)
            ymatrix.shape = [len(ymatrix), 1]
            label = "y" + grp
            if self.mcafit.STRIP:
                peaks_dict[label] = ymatrix + self.mcafit.zz
            else:
                peaks_dict[label] = ymatrix
            peaks_dict[label].shape = (len(peaks_dict[label]),)
            parameters[idx] = 0.0

        delimiter = ","
        header = '"channel"%s"Energy"%s"counts"%s"fit"%s"continuum"%s"pileup"' % (
            delimiter,
            delimiter,
            delimiter,
            delimiter,
            delimiter,
        )

        # add the peaks labels
        for key in peaks_dict:
            header += delimiter + ('"%s"' % key)
        # logging.getLogger("user_level_log").info("Writing %s" % fname)
        with open(fname, "w") as csv_fd:
            csv_fd.write(header)
            csv_fd.write("\n")
            for i in range(fitresult["result"]["xdata"].size):
                csv_fd.write(
                    "%.7g%s%.7g%s%.7g%s%.7g%s%.7g%s%.7g"
                    % (
                        fitresult["result"]["xdata"][i],
                        delimiter,
                        fitresult["result"]["energy"][i],
                        delimiter,
                        fitresult["result"]["ydata"][i],
                        delimiter,
                        fitresult["result"]["yfit"][i],
                        delimiter,
                        fitresult["result"]["continuum"][i],
                        delimiter,
                        fitresult["result"]["pileup"][i],
                    )
                )
                for key in peaks_dict:
                    csv_fd.write("%s%.7g" % (delimiter, peaks_dict[key][i]))

                csv_fd.write("\n")



"""

  File:  BIOMAXEigerMockup.py

  Description:  This module implements the hardware object for the Eiger detector
     based on a Tango device server


Detector Status:
-----------------

hardware status:
   ready:   ready for trigger (this is the state after an "Arm" command)
   idle:    ready for config (this should be the state after a "Disarm" command)

hardware object status:

   configuring:  a configuration task is ongoing


"""

import logging
import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import (
    cleanup,
    error_cleanup,
    task,
)


class BIOMAXEigerMockup(HardwareObject):
    """
    Description: Eiger hwobj based on tango
    """

    def __init__(self, *args):
        """
        Descrip. :
        """
        super().__init__(*args)

        self.device = None
        self.file_suffix = None
        self.default_exposure_time = 0.1
        self.default_compression = "bslz4"
        self.buffer_limit = None
        self.dcu = None
        self.config_state = None
        self.initialized = False
        self.status_chan = None
        self.roi_mode = "dsiabled"
        self.photon_energy = 12000
        self.energy_threshold = 6000

        # defaults
        self.energy_change_threshold_default = 20

    def init(self):

        tango_device = self.get_property("detector_device")
        filewriter_device = self.get_property("filewriter_device")

        self.file_suffix = self.get_property("file_suffix")
        self.default_exposure_time = self.get_property("default_exposure_time")
        self.default_compression = self.get_property("default_compression")
        self.buffer_limit = self.get_property("buffer_limit")
        self.dcu = self.get_property("dcu")

        # config needed to be set up for data collection
        # if values are None, use the one from the system
        self.col_config = {
            "OmegaStart": 0,
            "OmegaIncrement": 0.1,
            "BeamCenterX": 2000,  # length not pixel
            "BeamCenterY": 2000,
            "DetectorDistance": 0.15,
            "CountTime": 0.1,
            "NbImages": 100,
            "NbTriggers": 1,
            "ImagesPerFile": 100,
            "RoiMode": "disabled",
            "FilenamePattern": "test",
            "PhotonEnergy": 12000,
            "TriggerMode": "exts",
        }

        # we need to call the init device before accessing the channels here
        #   otherwise the initialization is triggered by the HardwareRepository Poller
        #   that is delayed after the application starts

        try:
            self.energy_change_threshold = float(
                self.get_property("min_trigger_energy_change")
            )
        except Exception:
            self.energy_change_threshold = self.energy_change_threshold_default

    def get_readout_time(self):
        return 0.000004

    def get_acquisition_time(self):
        return 2

    def get_roi_mode(self):
        return self.roi_mode

    def set_roi_mode(self, value):
        self.roi_mode = value

    def get_pixel_size_x(self):
        """
        return sizes of a single pixel along x-axis respectively
        unit, mm
        """
        x_pixel_size = 0.000075
        return x_pixel_size * 1000

    def get_pixel_size_y(self):
        """
        return sizes of a single pixel along y-axis respectively
        unit, mm
        """
        y_pixel_size = 0.000075
        return y_pixel_size * 1000

    def get_x_pixels_in_detector(self):
        """
        number of pixels along x-axis
        numbers vary depending on the RoiMode
        """
        return 4150

    def get_y_pixels_in_detector(self):
        """
        number of pixels along y-axis,
        numbers vary depending on the RoiMode
        """
        return 4371

    def get_minimum_exposure_time(self):
        return 0.01

    def get_sensor_thickness(self):
        return 0.45

    def has_shutterless(self):
        return True

    #  GET INFORMATION END

    #  SET VALUES
    def set_photon_energy(self, energy):
        """
        set photon_energy
        Note, the readout_time will be changed
        engery, in eV
        """
        self.photon_energy = energy
        return True

    def set_energy_threshold(self, threshold):
        """
        set energy_threshold
        Note, the readout_time will be changed
        By deafult, the value is 50% of the photon_energy and will be
        updated upon setting PhotonEnergy. If other values are needed,
        this should be set after changing PhotonEnergy.
        Eengery, in eV
        """
        self.energy_threshold = threshold

    #  SET VALUES END

    def prepare_acquisition(self, config):
        """
        config is a dictionary
        OmegaStart,OmegaIncrement,
        BeamCenterX
        BeamCenterY
        OmegaStart
        OmegaIncrement
        start, osc_range, exptime, ntrigger, number_of_images, images_per_file, compression,ROI,wavelength):
        """

        logging.getLogger("user_level_log").info("Preparing acquisition")

    @task
    def start_acquisition(self):
        """
        Before starting the acquisition a prepare_acquisition should be issued
        After prepare_acquisition the detector should be in "idle" state

        Otherwise you will have to send a "disarm" command by hand to be able to
        start an acquisition
        """

        logging.getLogger("user_level_log").info("Detector armed")

        return self.arm()

    def stop_acquisition(self):
        """
        when use external trigger, Disarm is required, otherwise the last h5 will
        not be released and not available in WebDAV.
        """

        try:
            self.cancel()  # this is needed as disarm in tango device server does not seem to work
            # as expected. the disarm command in the simpleinterface is always working
            # when called from Tango it does not. Once bug is solved in tango server, the
            # call to "cancel()" is not necessary here
            self.disarm()
        except Exception:
            pass

    def cancel_acquisition(self):
        """Cancel acquisition"""
        try:
            self.cancel()
        except Exception:
            pass

        time.sleep(1)
        self.disarm()

    def arm(self):
        return

    def trigger(self):
        return

    def disarm(self):
        return

    def cancel(self):
        return

    def abort(self):
        return



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
TangoLimaVideoDevice

[Note]
This class was previously called QtTangoLimaVideo.  It was changed to
TangoLimaVideoDevice as it can be used now in Qt4(Qt5) environment or
without it.
The property <useqt>False</useqt> in the xml configuration file will inhibit
the use of qt.

The Hardware Object will poll for images at regular interval
(set by property <interval>value-in-msecs</interval>
For every image received a signal "imageReceived" is emitted. The content of the
information emitted varies:
- When qt is used:
   a "qimage" is delivered as parameter with the signal

- When qt is not used  (for web applications):
   three values are sent with the signal
     val1:  string with jpeg encoding of the image
     val2:  width (pixels) of the image
     val3:  height (pixels) of the image

   (web, non-qt encoding, has been validated for Prosilica cameras
     but in principle other cameras could also adopt the same mechanism)

[Description]
HwObj used to grab images via Tango Lima device server
If you want to access the Lima Library directly you may consider using
the LimaVideoDevice module instead

[Configuration]
Example Hardware Object XML file :
==================================
<object class="QtLimaVideo">
   <type>basler</type>
   <encoding>yuv422p</encoding>
   <tangoname>bl13/eh/lima_oav</tangoname>
   <gain>0.5</gain>
   <exposure>0.01</exposure>
   <mirror>(False, False)</mirror>
   <interval>30</interval>
</object>
"""
from __future__ import print_function

import struct

import numpy as np
import PyTango

from mxcubecore.HardwareObjects.abstract.AbstractVideoDevice import AbstractVideoDevice


class TangoLimaVideoDevice(AbstractVideoDevice):
    """
    Descript. :
    """

    def __init__(self, name):
        """
        Descript. :
        """
        AbstractVideoDevice.__init__(self, name)
        self.device = None

    def init(self):
        """
        Descript. :
        """

        tangoname = self.get_property("tangoname")
        endian = self.get_property("endian")

        if endian in ["small", "Small", "Little", "little"]:
            self.header_fmt = "<IHHqiiHHHH"
        else:
            self.header_fmt = ">IHHqiiHHHH"

        self.header_size = struct.calcsize(self.header_fmt)

        self.device = PyTango.DeviceProxy(tangoname)
        self.device.ping()

        AbstractVideoDevice.init(self)

    def set_cam_encoding(self, cam_encoding):
        if cam_encoding == "yuv422p":
            self.device.video_mode = "YUV422"
        elif cam_encoding == "y8":
            self.device.video_mode = "Y8"

        AbstractVideoDevice.set_cam_encoding(self, cam_encoding)

    """ Overloading of AbstractVideoDevice methods """

    def get_image_dimensions(self):
        return [self.device.image_width, self.device.image_height]

    def get_image(self):
        """
        Reads image from `video_last_image` attribute of lima device proxy,
        which is type of `bytes` and converts it into np.array of int.

        Returns
            raw_buffer : 1d np.array of np.uint16
                Image
            width : int
                Image width
            height : int
                Image height
        """
        img_data = self.device.video_last_image

        if img_data[0] == "VIDEO_IMAGE":
            _ = img_data[1][: self.header_size]
            raw_buffer = np.fromstring(img_data[1][self.header_size :], np.uint16)
            _, _, _, _, width, height, _, _, _, _ = struct.unpack(
                self.header_fmt, img_data[1][: self.header_size]
            )
            return raw_buffer, width, height
        else:
            return None, 0, 0

    def get_gain(self):
        if self.get_cam_type() == "basler":
            value = self.device.video_gain
            return value

    def set_gain(self, gain_value):
        if self.get_cam_type() == "basler":
            self.device.video_gain = gain_value
            return

    def get_exposure_time(self):
        if self.get_cam_type() == "basler":
            return self.device.video_exposure

    def set_exposure_time(self, exposure_time_value):
        if self.get_cam_type() == "basler":
            self.device.video_exposure = exposure_time_value

    def get_video_live(self):
        return self.device.video_live

    def set_video_live(self, flag):
        self.device.video_live = flag

    """ END Overloading of AbstractVideoDevice methods """


def test_hwo(hwo):
    print("Image dimensions: ", hwo.get_image_dimensions())
    print("Live Mode: ", hwo.get_video_live())



import os
import subprocess

from devtools import debug
from pydantic.v1 import (
    BaseModel,
    Field,
)
from typing_extensions import Literal

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.common import (
    CommonCollectionParamters,
    LegacyParameters,
    PathParameters,
    StandardCollectionParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class TestUserCollectionParameters(BaseUserCollectionParameters):
    class Config:
        extra: "ignore"


class TestCollectionTaskParameters(SsxBaseQueueTaskParameters):
    path_parameters: PathParameters
    common_parameters: CommonCollectionParamters
    collection_parameters: StandardCollectionParameters
    user_collection_parameters: TestUserCollectionParameters
    legacy_parameters: LegacyParameters

    @staticmethod
    def update_dependent_fields(field_data):
        new_data = {"exp_time": field_data["sub_sampling"] * 2}
        return new_data


class TestCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class TestCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = TestCollectionQueueModel
    DATA_MODEL = TestCollectionTaskParameters
    NAME = "TestCollection"
    REQUIRES = ["point", "line", "no_shape", "chip", "mesh"]

    def __init__(self, view, data_model: TestCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()
        debug(self._data_model._task_data)

        data_root_path = HWR.beamline.session.get_image_directory(
            os.path.join(
                self._data_model._task_data.path_parameters.subdir,
                self._data_model._task_data.path_parameters.experiment_type,
            )
        )

        process_path = os.path.join(
            HWR.beamline.session.get_base_process_directory(),
            self._data_model._task_data.path_parameters.subdir,
        )

        subprocess.Popen(
            "mkdir --parents %s" % (data_root_path),
            shell=True,
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        ).wait()

        dcg = HWR.beamline.lims.pyispyb.create_ssx_data_collection_group()
        HWR.beamline.lims.pyispyb.create_ssx_data_collection(dcg)

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



"""Class for cameras connected to Lima Tango Device Servers

Example configuration:
----------------------
<object class="TangoLimaVideo">
  <username>Prosilica 1350C</username>
  <tangoname>id23/limaccd/minidiff2</tangoname>
  <bpmname>id23/limabeamviewer/minidiff2</bpmname>
  <interval>15</interval>
  <video_mode>RGB24</video_mode>
</object>

If video mode is not specified, BAYER_RG16 is used by default.
"""

import io
import logging
import struct
import time

import gevent
import PyTango
from PIL import Image
from PyTango.gevent import DeviceProxy

from mxcubecore import BaseHardwareObjects


def poll_image(lima_tango_device, video_mode, FORMATS):
    img_data = lima_tango_device.video_last_image

    hfmt = ">IHHqiiHHHH"
    hsize = struct.calcsize(hfmt)
    _, _, _, _, width, height, _, _, _, _ = struct.unpack(hfmt, img_data[1][:hsize])

    raw_data = img_data[1][hsize:]
    _from, _to = FORMATS.get(video_mode, (None, None))

    if _from and _to:
        img = Image.frombuffer(_from, (height, width), raw_data, "raw", _from, 0, 1)

        img_bytes = io.BytesIO()
        img.save(img_bytes, format=_to)
        img = img.tobytes()
    else:
        img = raw_data

    return img, width, height


class TangoLimaVideo(BaseHardwareObjects.HardwareObject):
    def __init__(self, name):
        super().__init__(name)
        self.__polling = None
        self._video_mode = None
        self._last_image = (0, 0, 0)

        # Dictionary containing conversion information for a given
        # video_mode. The camera video mode is the key and the first
        # index of the tuple contains the corresponding PIL mapping
        # and the second the desried output format. The image is
        # passed on as it is of the video mode is not in the dictionary
        self._FORMATS = {
            "RGB8": ("L", "BMP"),
            "RGB24": ("RGB", "BMP"),
            "RGB32": ("RGBA", "BMP"),
            "NO_CONVERSION": (None, None),
        }

    def init(self):
        self.device = None

        try:
            self._video_mode = self.get_property("video_mode", "RGB24")
            self.device = DeviceProxy(self.tangoname)
            # try a first call to get an exception if the device
            # is not exported
            self.device.ping()
        except PyTango.DevFailed as traceback:
            last_error = traceback[-1]
            logging.getLogger("HWR").error("%s: %s", str(self.name()), last_error.desc)

            self.device = BaseHardwareObjects.Null()
        else:
            if self.get_property("exposure_time"):
                self._sleep_time = float(self.get_property("exposure_time"))
            elif self.get_property("interval"):
                self._sleep_time = float(self.get_property("interval")) / 1000.0

            if self.get_property("control_video", "True"):
                logging.getLogger("HWR").info("MXCuBE controlling video")

                if self.device.video_live:
                    self.device.video_live = False

                self.device.video_mode = self._video_mode
                self.set_exposure(self._sleep_time)

                self.device.video_live = True
            else:
                logging.getLogger("HWR").info("MXCuBE NOT controlling video")

        self.update_state(BaseHardwareObjects.HardwareObjectState.READY)

    def get_last_image(self):
        return poll_image(self.device, self.video_mode, self._FORMATS)

    def _do_polling(self, sleep_time):
        lima_tango_device = self.device

        while True:
            data, width, height = poll_image(
                lima_tango_device, self.video_mode, self._FORMATS
            )

            self._last_image = data, width, height
            self.emit("imageReceived", data, width, height, False)
            time.sleep(sleep_time)

    def connect_notify(self, signal):
        if signal == "imageReceived":
            if self.__polling is None:
                self.__polling = gevent.spawn(
                    self._do_polling, self.device.video_exposure
                )

    def get_width(self):
        return self.device.image_width

    def get_height(self):
        return self.device.image_height

    def set_live(self, mode):
        curr_state = self.device.video_live

        if mode:
            if not curr_state:
                self.device.video_live = True
        else:
            if curr_state:
                self.device.video_live = False

    def set_exposure(self, exposure):
        self.device.video_exposure = exposure



from mxcubecore.BaseHardwareObjects import HardwareObject

from .BeamCmds import (
    ControllerCommand,
    HWObjActuatorCommand,
)


class ID231BeamCmds(HardwareObject):
    """BeamCmds for ID23-1"""
    def __init__(self, *args):
        HardwareObject.__init__(self, *args)

    def init(self):
        controller = self.get_object_by_role("controller")
        detcover = self.get_object_by_role("detcover")
        scintilator = self.get_object_by_role("scintilator")
        aperture = self.get_object_by_role("aperture")
        hutchtrigger = self.get_object_by_role("hutchtrigger")
        cryo = self.get_object_by_role("cryostream")

        self.centrebeam = ControllerCommand("Centre beam", controller.centrebeam)
        self.quick_realign = ControllerCommand(
            "Quick realign", controller.quick_realign
        )
        # self.anneal = ControllerCommand("Anneal", controller.anneal_procedure)

        self.detcover = HWObjActuatorCommand("Detector cover", detcover)
        self.scintilator = HWObjActuatorCommand("Scintillator", scintilator)
        self.aperture = HWObjActuatorCommand("Aperture", aperture)
        self.hutchtrigger = HWObjActuatorCommand("Hutchtrigger", hutchtrigger)
        self.cryo = HWObjActuatorCommand("Cryostream", cryo)

    def get_commands(self):
        """Define the commands to be used"""
        return [
            self.centrebeam,
            self.quick_realign,
            # self.anneal,
            self.detcover,
            self.scintilator,
            self.aperture,
            self.hutchtrigger,
            self.cryo,
        ]



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""FlexHarvester Linux Java implementation of the Flex sample changer for Harvester Use.
Example xml file:
<object class = "EMBLFlexHarvester">
  <username>Sample Changer</username>
  <exporter_address>lid231flex1:9001</exporter_address>
</object>
"""
from __future__ import annotations

import logging
import time
from typing import (
    Any,
    List,
)

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore import queue_entry
from mxcubecore.HardwareObjects.EMBLFlexHCD import EMBLFlexHCD
from mxcubecore.model import queue_model_objects as qmo
from mxcubecore.TaskUtils import task


class EMBLFlexHarvester(EMBLFlexHCD):
    __TYPE__ = "Flex Sample Changer"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._harvester_hwo = None
        self.pin_cleaning = None
        self._loaded_sample = None

    def init(self):
        self.pin_cleaning = self.get_property("pin_cleaning")

        self._loaded_sample = (-1, -1, -1)
        self._harvester_hwo = self.get_object_by_role("harvester")

        EMBLFlexHCD.init(self)

    def get_room_temperature_mode(self):
        return self._execute_cmd_exporter("getRoomTemperatureMode", attribute=True)

    def set_room_temperature_mode(self, value):
        self._execute_cmd_exporter("setRoomTemperatureMode", value, command=True)
        logging.getLogger("user_level_log").info(
            f"setting Robot Room temperature to {value}"
        )
        return self.get_room_temperature_mode()

    def mount_from_harvester(self):
        return True

    def get_sample_list(self) -> List[Any]:
        """
        Get Sample List related to the Harvester content/processing Plan
        """
        sample_list = super().get_sample_list()
        present_sample_list = []
        ha_sample_lists = self._harvester_hwo.get_crystal_uuids()
        ha_sample_names = self._harvester_hwo.get_sample_names()
        ha_sample_acronyms = self._harvester_hwo.get_sample_acronyms()

        if ha_sample_lists:
            for i in range(len(ha_sample_lists)):
                sample = sample_list[i]
                sample.id = ha_sample_lists[i]
                sample._name = ha_sample_names[i]
                # if all sample come with proteinAcronym
                if len(ha_sample_acronyms) > 0 and len(ha_sample_acronyms) == len(
                    ha_sample_lists
                ):
                    sample.proteinAcronym = ha_sample_acronyms[i]
                else:
                    # if all sample does not have proteinAcronym
                    # we set first proteinAcronym to all if exist at least one
                    sample.proteinAcronym = (
                        ha_sample_acronyms[0] if len(ha_sample_acronyms) > 0 else ""
                    )
                present_sample_list.append(sample)

        return present_sample_list

    def _hw_get_mounted_sample(self) -> str:
        loaded_sample = self._loaded_sample
        return (
            str(loaded_sample[0])
            + ":"
            + str(loaded_sample[1])
            + ":"
            + "%02d" % loaded_sample[2]
        )

    @task
    def load_a_pin_for_calibration(self):
        """
        Load a Pin from Harvester
        """
        try:
            self.prepare_load()
            self.enable_power()

            load_task = gevent.spawn(
                self._execute_cmd_exporter,
                "loadSampleFromHarvester",
                self.pin_cleaning,
                command=True,
            )

            self._wait_busy(30)
            err_msg = "Timeout while waiting to sample to be loaded"
            with gevent.Timeout(600, RuntimeError(err_msg)):
                while not load_task.ready():
                    gevent.sleep(2)

            with gevent.Timeout(600, RuntimeError(err_msg)):
                while True:
                    is_safe = self._execute_cmd_exporter(
                        "getRobotIsSafe", attribute=True
                    )
                    if is_safe:
                        break
                    gevent.sleep(2)
            return True
        except Exception:
            return False

    def start_harvester_centring(self):
        try:
            dm = HWR.beamline.diffractometer

            logging.getLogger("user_level_log").info("Start Auto Harvesting Centring")

            computed_offset = HWR.beamline.harvester.get_offsets_for_sample_centering()
            dm.start_harvester_centring(computed_offset)

        except Exception:
            logging.getLogger("user_level_log").exception(
                "Could not center sample, skipping"
            )
            raise queue_entry.QueueSkipEntryException(
                "Could not center sample, skipping", ""
            )

    def _set_loaded_sample_and_prepare(self, loaded_sample_tup, previous_sample_tup):
        res = False

        loaded_sample = self.get_sample_with_address(loaded_sample_tup)

        if -1 not in loaded_sample_tup and loaded_sample_tup != previous_sample_tup:
            self._set_loaded_sample(loaded_sample)
            self._prepare_centring_task()
            res = True

        if res:
            if self._harvester_hwo.get_room_temperature_mode() == False:
                self.queue_harvest_next_sample(loaded_sample.get_address())

            # in this case we expect CENTRING_METHOD=None
            self.start_harvester_centring()

        return res

    def _do_load(self, sample=None):
        """
        Load a Sample from Harvester
        """
        self.queue_harvest_sample(sample.get_address())
        self._update_state()

        # We wait for the sample changer if its already doing something, like defreezing
        # wait for 10 minutes then timeout !
        self._wait_ready(600)

        previous_sample = self._loaded_sample
        # Start loading from harvester
        load_task = gevent.spawn(
            self._execute_cmd_exporter,
            "loadSampleFromHarvester",
            self.pin_cleaning,
            command=True,
        )

        # Wait for sample changer to start activity
        try:
            _tt = time.time()
            self._wait_busy(300)
            logging.getLogger("HWR").info(f"Waited SC activity {time.time() - _tt}")
        except Exception:
            logging.getLogger("user_level_log").error(
                "ERROR While Waited SC activity to start"
            )
            for msg in self.get_robot_exceptions():
                logging.getLogger("user_level_log").error(msg)
            raise

        # Wait for the sample to be loaded, (put on the goniometer)
        err_msg = "Timeout while waiting to sample to be loaded"
        with gevent.Timeout(600, RuntimeError(err_msg)):
            while not load_task.ready():
                gevent.sleep(2)

        with gevent.Timeout(600, RuntimeError(err_msg)):
            while True:
                is_safe = self._execute_cmd_exporter("getRobotIsSafe", attribute=True)

                if is_safe:
                    break

                gevent.sleep(2)

        for msg in self.get_robot_exceptions():
            if msg is not None:
                logging.getLogger("user_level_log").error(
                    "ERROR While SC activity After Loaded Sample "
                )
                logging.getLogger("HWR").error(msg)
                logging.getLogger("user_level_log").error(msg)
                # Temp: In Harvester mode any robot Exception is consider as Loading failed
                # Except Pin Cleaning Station Exception
                if "Pin Cleaning Station" not in msg:
                    return False

        loaded_sample = (
            sample.get_cell_no(),
            sample.get_basket_no(),
            sample.get_vial_no(),
        )
        self._loaded_sample = loaded_sample

        return self._set_loaded_sample_and_prepare(loaded_sample, previous_sample)

    def harvest_and_mount_sample(self, xtal_uuid: str, sample) -> bool:
        """
        return (Bool) : wether the crystal has been Harvest then mount

        """
        try:
            self._harvester_hwo.harvest_crystal(xtal_uuid)
            self._harvester_hwo._wait_sample_transfer_ready(None)

            self._do_load(sample)
        except Exception:
            logging.getLogger("user_level_log").exception("Could not Harvest Crystal")
            return "Could not Harvest Crystal"

    def queue_list(self) -> List[str]:
        """
        builds a List representation of the queue based.
        """

        node = HWR.beamline.queue_model.get_model_root()

        result = []

        if isinstance(node, List):
            node_list = node
        else:
            node_list = node.get_children()

        for node in node_list:
            if isinstance(node, qmo.Sample):
                result.append(node.loc_str)

        return result

    def get_sample_uuid(self, sampleID: str) -> str:
        samples_list = self.get_sample_list()
        sample_uuid = None
        for s in samples_list:
            if s.get_address() == sampleID or s.get_id() == sampleID:
                sample_uuid = s.get_id()
                return sample_uuid

    def queue_harvest_sample(self, sample_loc_str) -> None:
        """
        While queue execution send harvest request
        """
        current_queue = self.queue_list()

        sample_uuid = self.get_sample_uuid(sample_loc_str)

        self._harvester_hwo.queue_harvest_sample(
            sample_loc_str, sample_uuid, current_queue
        )

    def queue_harvest_next_sample(self, sample_loc_str) -> None:
        """
        While queue execution send harvest request
        on next sample of the queue list
        """

        current_queue_list = self.queue_list()

        next_sample = None
        try:
            next_sample = current_queue_list[
                current_queue_list.index(sample_loc_str) + 1
            ]
        except (ValueError, IndexError):
            next_sample = None

        sample_uuid = self.get_sample_uuid(next_sample)

        self._harvester_hwo.queue_harvest_next_sample(next_sample, sample_uuid)
